# vi: ts=4 sw=4 et
#
# application.py: the global Application object
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: application.py,v 1.25 2003/03/18 19:42:01 geertj Exp $

from draco.namespace import NamespaceBridge
from draco.database import database
from draco.server import server


class ApplicationNamespace(NamespaceBridge):
    """
    The "application" namespace. This is a namespace that is bound to the
    specific "run" of the application. If the applications starts or
    restarts, this namespace is reinitialized.

    Each application namespace has a scope. Namespaces with different
    scopes are independent.
    """

    def setScope(self, generation, scope):
        if generation is not None:
            generation += 1
        namespace = database.createNamespace('draco_application_data', \
                    'generation', generation, scope)
        self.setNamespace(namespace)


class Application(ApplicationNamespace):
    """
    Application object. This object handles application global functionality.
    This includes:

    * The default application namespace (scope = '__system__').
    * locking

    Database acccess needs to be configured to use Application.
    """

    def configure(self, config):
        pass

    def reload(self):
        """
        Reload the application object.
        """
        self.m_generation = server.generation()
        self.setScope(self.m_generation, '__system__')

    def __nonzero__(self):
        """
        Return nonzero if the application namespaces are available.
        """
        return self.available()

    def namespace(self, scope):
        """
        Return the application namespace with scope `scope'.
        """
        namespace = ApplicationNamespace()
        namespace.setScope(self.m_generation, scope)
        return namespace

    def lock(self):
        """
        Lock the appliation.
        """
        database.lock()

    def unlock(self):
        """
        Unlock the application.
        """
        database.unlock()
