# vi: ts=4 sw=4 et
#
# config.py: configuration subsystem
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: config.py,v 1.8 2003/03/17 11:22:43 geertj Exp $

import re
import os.path

import draco
from draco.namespace import DictNamespace
from draco.filewatch import FileWatch


class ConfigNamespace(DictNamespace):
    """
    The "configuration" namespace. This is a namespace that corresponds
    with a section from the draco config file. A configuration namespace
    is read-only.
    """

    def __init__(self):
        DictNamespace.__init__(self)


class Config(ConfigNamespace):
    """
    The Config object. This objects parses draco's configuration files
    and provides access to the configuration namespaces.
    """

    re_ignore = re.compile('^\s*(#.*)?$')
    re_section = re.compile('^\s*\[\s*(\w+)\s*\]\s*$')
    re_assign = re.compile('^\s*(\w+)\s*=\s*(.*)$')

    def __init__(self, files):
        """
        Initialize the configuration subsystem. FILES must be a sequence
        of filenames to parse.
        """
        ConfigNamespace.__init__(self)
        self.m_configs = files
        self.m_filewatch = FileWatch(files)
        self.m_modified = 0

    def reload(self, options):
        """
        Reload the config object.
        """
        if not self.m_filewatch.modified():
            self.m_modified = 0
            return
        self.clear()
        self.m_sections = {}
        for name in options.keys():
            if name == 'DocumentRoot':
                value = os.path.normpath(options[name])
            else:
                value = options[name]
            self[name] = value
        for i in self.m_configs:
            self.parse(i)
        self.m_modified = 1

    def parse(self, fname):
        """
        Parse the configuration file FNAME.
        """
        try:
            fconfig = file(fname)
        except IOError:
            return
        section = '__system__'
        lineno = 0
        for line in fconfig:
            lineno += 1
            if self.re_ignore.match(line):
                continue
            match = self.re_section.match(line)
            if match:
                section = match.group(1)
                self.m_sections[section] = ConfigNamespace()
                continue
            match = self.re_assign.match(line)
            if not match:
                self.m_filewatch.clear(fname)
                raise draco.HttpInternalServerError, \
                        'Syntax error in configure file %s:%d' % (fname, lineno)
            name = match.group(1)
            try:
                value = eval(match.group(2))
            except:
                self.m_filewatch.clear(fname)
                raise draco.HttpInternalServerError, \
                        'Syntax error in configuraton file %s:%d' % (fname, lineno)
            if section == '__system__':
                self[name] = value
            else:
                self.m_sections[section][name] = value

    def modified(self):
        """
        Returns nonzero if the last reload() call modified the config.
        """
        return self.m_modified

    def namespace(self, scope):
        """
        Return the configuration namespace SCOPE.
        """
        try:
            return self.m_sections[scope]
        except KeyError:
            return ConfigNamespace()
