# vi: ts=4 sw=4 et
#
# cookies.py: the cookie namespace
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: cookie.py,v 1.6 2003/03/10 16:38:37 geertj Exp $

from draco.logger import logger
from draco.namespace import DictNamespace
from draco.date import DateTime


class Cookie:
    """
    A single cookie with its attributes.
    """

    def __init__(self, name, value, path='/', domain=None,  expires=None,
                 secure=None):
        self.name = name
        self.value = value
        self.path = path
        self.domain = domain
        self.expires = expires
        self.secure = secure

    def __str__(self):
        s = '%s=%s' % (self.name.encode('form'), self.value.encode('form'))
        s += '; Path=%s' % self.path
        if self.domain:
            s += '; Domain=%s' % self.domain
        if self.expires:
            s += '; Expires=%s' % self.expires.format('nscookie')
        if self.secure:
            s += '; Secure'
        return s


class Cookies(DictNamespace):
    """
    This class implements the global "cookies" namespaces.
    """

    def __init__(self):
        DictNamespace.__init__(self)
        self.m_path = '/'
        self.m_domain = None
        self.m_expires = None
        self.m_secure = None

    def configure(self, config):
        pass

    def reload(self):
        self.clear()
        self.m_update = {}

    def parseCookies(self, request):
        """
        Parse the http "Cookie" header.
        """
        header = request.header('Cookie')
        if not header:
            return
        cookies = {}
        try:
            parts = header.split(';')
            for pa in parts:
                name, value = map(lambda x: x.strip().decode('form'), pa.split('='))
                if name.startswith('$'):
                    continue
                cookies[name] = value
        except:
            logger.warn('Illegal cookie format: %s' % header)
            return
        self.update(cookies)
        self.m_update = {}

    def setCookies(self, response):
        """
        Set the "Set-Cookie" header in the response.
        """
        for co in self.m_update.values():
            response.addHeader('Set-Cookie', str(co))

    def __setitem__(self, name, value):
        if type(value) != str:
            value = str(value)
        DictNamespace.__setitem__(self, name, value)
        self.m_update[name] = Cookie(name, value, self.m_path, self.m_domain,
                                     self.m_expires, self.m_secure)

    def __delitem__(self, name):
        """
        Delete a cookie NAME.
        """
        DictNamespace.__delitem__(self, name)
        self.m_update[name] = Cookie(name, '', self.m_path, self.m_domain,
                                     DateTime(0), self.m_secure)

    def setPath(self, name, path):
        """
        Set the "Path" attribute for cookie NAME to PATH.
        """
        try:
            cookie = self.m_update[name]
        except KeyError:
            cookie = Cookie(name, self[name], self.m_path, self.m_domain,
                            self.m_expires, self.m_secure)
            self.m_update[name] = cookie
        cookie.path = path

    def setDefaultPath(self, path):
        """
        Set the default "Path" attribute.
        """
        self.m_path = path

    def setDomain(self, name, domain):
        """
        Set the "Domain" attribute for cookie NAME to DOMAIN.
        """
        try:
            cookie = self.m_update[name]
        except KeyError:
            cookie = Cookie(name, self[name], self.m_path, self.m_domain,
                            self.m_expires, self.m_secure)
            self.m_update[name] = cookie
        cookie.domain = domain

    def setDefaultDomain(self, domain):
        """
        Set the default "Domain" attribute.
        """
        self.m_domain = domain

    def setExpires(self, name, expires):
        """
        Set the "Expires" attribute for cookie NAME to EXPIRES. EXPIRES
        must be a DateTime object.
        """
        try:
            cookie = self.m_update[name]
        except KeyError:
            cookie = Cookie(name, self[name], self.m_path, self.m_domain,
                            self.m_expires, self.m_secure)
            self.m_update[name] = cookie
        cookie.expires = expires

    def setDefaultExpires(self, expires):
        """
        Set the default "Expires" attribute. EXPIRES must be a DateTime object.
        """
        self.m_expires = expires

    def setSecure(self, name, secure):
        """
        Set the "Secure" option for a cookie NAME if SECURE is nonzero.
        """
        try:
            cookie = self.m_update[name]
        except KeyError:
            cookie = Cookie(name, self[name], self.m_path, self.m_domain,
                            self.m_expires, self.m_secure)
            self.m_update[name] = cookie
        cookie.secure = secure

    def setDefaultSecure(self, secure):
        """
        Set the "Secure" attribute by default if SECURE is nonzero.
        """
        self.m_secure = secure
