# vi: ts=4 sw=4 et
#
# database.py: database access
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: database.py,v 1.9 2003/03/18 13:56:25 geertj Exp $

import cPickle

from draco.namespace import Namespace


class Database:
    """
    Abstract base class for all databases.
    """

    def configure(self, config):
        raise NotImplementedError

    def reload(self):
        raise NotImplementedError

    def finalize(self):
        raise NotImplementedError

    def checkTables(self):
        raise NotImplementedError

    def createNamespace(self, table, discr, value, scope):
        raise NotImplementedError

    def connection(self):
        """
        Return the global database connection.
        """
        raise NotImplementedError

    def cursor(self):
        """
        Return the global database cursor.
        """
        raise NotImplementedError

    def commit(self):
        """
        An alias for connection().commit().
        """
        self.connection().commit()

    def rollback(self):
        """
        An alias for connection().rollback().
        """
        self.connection().rollback()

    def lock(self):
        """
        Lock the whole database. This is used for application.lock().
        """
        raise NotImplementedError

    def unlock(self):
        """
        Unlock the database.
        """
        raise NotImplementedError


class DatabaseFactory:
    """
    The database factory can create a specific database instance according to
    the user configuration.
    """

    def configure(self, config):
        """
        Configure the database factory.
        """
        database = config.get('Database')
        if not database:
            database = 'null'
        elif isinstance(database, int):
            database = 'mysql'
        else:
            database = database.lower()
        self.m_database = database

    def createDatabase(self):
        """
        Create a new database instance according to the configuration.
        """
        if self.m_database == 'mysql':
            from draco.dbmysql import MySQLDatabase
            db = MySQLDatabase()
        elif self.m_database == 'pgsql':
            from draco.dbpgsql import PgSQLDatabase
            db = PgSQLDatabase()
        else:
            from draco.dbnull import NullDatabase
            db = NullDatabase()
        return db

    def createDbapi(self):
        """
        Return the database api for the configured database.
        """
        if self.m_database == 'mysql':
            from draco.dbmysql import dbapi
        elif self.m_database == 'pgsql':
            from draco.dbpgsql import dbapi
        else:   
            from draco.dbnull import dbapi
        return dbapi


class DatabaseNamespace(Namespace):
    """
    A database namespace is a namespace with serialize()/deserialize()
    methods, an available() method and some properties that define which
    table to use.
    """

    def __init__(self, database, table, key, value, scope):
        Namespace.__init__(self)
        self.database = database
        self.table = table
        self.key = key
        self.value = value
        self.scope = scope

    def serialize(self, value):
        return cPickle.dumps(value, 1)

    def deserialize(self, value):
        return cPickle.loads(value)

    def available(self):
        raise NotImplementedError
