# vi: ts=4 sw=4 et
#
# date.py: the Date and DateTime objects
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: date.py,v 1.42 2003/03/25 11:45:40 geertj Exp $

import time


class Date(object):
    """
    An object to work work with dates in the Gregorian calendar. It supports
    reading various input formats, formatting to various output formats and
    basic arithmetic (add, sub, compare).

    Internally the date is kepts as a (year, month, day) tuple. The range of
    possible dates is, in principle, only limited by their storage type.

    In practice however, this is of limited use because the Gregorian calendar
    was not always in effect. In many catholic countries, the Gregorian 
    calendar was installed in 1582, but in other countries (like Russia), this 
    did not happen until 1918.
    """

    _monthoffset = ( 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 )
    _formats = { 'mysql': '%Y-%m-%d', 'ctime': '%a %b %d %Y',
                 'iso8601': '%Y-%m-%d' }
    _nullfmt = '0000-00-00'


    def __init__(self, *args):
        """
        Initialize the Date object. The constructor takes either zero or one
        argument. If no argument is passed the Date is initialized to the
        current date. If an argument is given, it must be a null date (None or
        the literal string '0000-00-00), a string in a supported format or an
        object with a tuple() method (like an mx.DateTime instance).
        """
        self.m_formats = self._formats.copy()
        if len(args) == 0:
            self.m_tm = list(time.localtime())
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, Date):
                self.m_tm = arg.m_tm[:]
            elif arg in (None, self._nullfmt):
                self.m_tm = [0]*9
            elif type(arg) in (str, unicode):
                self.parse(args[0])
            elif hasattr(arg, 'tuple'):
                self.m_tm = map(int, arg.tuple())
            elif hasattr(arg, 'timetuple'):
                self.m_tm = list(arg.timetuple())
            else:
                raise TypeError, 'Unsupported date type.'
        else:
            raise TypeError, 'Expecting 0 or 1 arguments.'
        self.m_tm[3:6] = [0]*3

    def parse(self, str):
        """
        Parse the date string STR and set the current date.
        """
        parsed = None
        for fmt in self.m_formats.values():
            try:
                parsed = list(time.strptime(str, fmt))
                break
            except ValueError:
                pass
        if not parsed:
            raise ValueError, 'Unknown date format.'
        self.m_tm = parsed

    def copy(self):
        """
        Return a copy of the date.
        """
        cls = self.__class__
        return cls(self)

    def __nonzero__(self):
        return self.m_tm != [0]*9

    def __repr__(self):
        return self.format('mysql')

    __str__ = __repr__

    def __cmp__(self, other):
        if isinstance(other, Date):
            return cmp(self.m_tm[:6], other.m_tm[:6])
        else:
            return cmp(id(self), id(other))

    def _get_year(self):
        return self.m_tm[0]

    def _set_year(self, year):
        if type(year) != int:
            raise ValueError, 'Illegal year %s' % year
        self.m_tm[0] = year
        self.addDate(0,0,0)  # renormalize

    year = property(_get_year, _set_year, None)

    def _get_month(self):
        return self.m_tm[1]

    def _set_month(self, month):
        if type(month) != int:
            raise ValueError, 'Illegal month %s' % month
        self.m_tm[1] = month
        self.addDate(0,0,0)

    month = property(_get_month, _set_month, None)

    def _get_day(self):
        return self.m_tm[2]

    def _set_day(self, day):
        if type(day) != int:
            raise ValueError, 'Illegal day %s' % day
        self.m_tm[2] = day
        self.addDate(0,0,0)

    day = property(_get_day, _set_day, None)

    def _get_weekday(self):
        return self.m_tm[6]

    weekday = property(_get_weekday, None, None)

    def _get_yearday(self):
        return self.m_tm[7]

    yearday = property(_get_yearday, None, None)

    def _get_monthdays(self):
        return self._monthdays(self.m_tm[0], self.m_tm[1])

    monthdays = property(_get_monthdays, None, None)

    def _get_isleap(self):
        return self._isleap(self.m_tm[0])

    isleap = property(_get_isleap, None, None)

    def _weekday(self, year, month, day):
        """
        Return the day of week of date (year, month, day). Monday is zero.
        """
        leaps = (year-1)/4 - (year-1)/100 + (year-1)/400
        dow = year*365 + self._monthoffset[month-1] + day + leaps + 5
        if self._isleap(year) and month > 2:
            dow += 1
        return dow%7

    def _yearday(self, year, month, day):
        """
        Return the year day (julian day) of the date (year, month, day).
        """
        jday = self._monthoffset[month-1] + day
        if self._isleap(year) and month > 2:
            jday += 1
        return jday
        
    def _isleap(self, year):
        """
        Return nonzero if YEAR is leap.
        """
        leap = not year%4 and (year%100 or not year%400)
        return leap

    def _monthdays(self, year, month):
        """
        Return the number of months in MONTH of YEAR.
        """
        if month in (1, 3, 5, 7, 8, 10, 12): 
            return 31
        elif month in (4, 6, 9, 11): 
            return 30
        elif self._isleap(year): 
            return 29
        else:
            return 28

    def addDate(self, dy, dm, dd):
        """
        Add DY years, DM months and DD days to the current date. The years
        are added first, then the months and finally the days. Order does
        matter here because not every month and year have the same number
        of days.
        """
        if not self:
            return
        year, month, day = self.m_tm[:3]
        year += dy
        month += dm
        if not 1 <= month <= 12:
            dy, month = divmod(month-1, 12)
            year += dy
            month += 1

        # Split out periods of 400 years in case we're adding huge
        # amount of days. Every period of 400 years has the same number
        # of days.

        if not -146097 < dd < 146097:  # 400 years
            dy, dd = divmod(dd, 146097)
            year += dy*400

        # Add the days and normalize them one month at a time.

        day += dd
        while day > self._monthdays(year, month):
            day -= self._monthdays(year, month)
            month += 1
            if month == 13:
                month = 1
                year += 1
        while day < 1:
            month -= 1
            if month == 0:
                month = 12
                year -= 1
            day += self._monthdays(year, month)

        self.m_tm[:3] = [year, month, day]
        self.m_tm[6] = self._weekday(year, month, day)
        self.m_tm[7] = self._yearday(year, month, day)

    def strftime(self, format):
        """
        Format the date using a strftime() format string FORMAT.
        """
        if self:
            return time.strftime(format, self.m_tm)
        else:
            return ''

    def format(self, format):
        """
        Formats the date using a predefined format FORMAT.
        """
        try:
            return self.strftime(self.m_formats[format])
        except KeyError:
            raise ValueError, 'No such date format.'

    def setFormat(self, name, value):
        """
        Set the format NAME to VALUE.
        """
        self.m_formats[name] = value

    def setDefaultFormat(cls, name, value):
        """
        Set the default format NAME to VALUE. VALUE must be a time.strftime compatible
        format string.
        """
        cls._formats[name] = value

    setDefaultFormat = classmethod(setDefaultFormat)


class DateTime(Date):
    """
    DateTime: A date/time handling class. It supports reading various input 
    formats, formatting to various output formats and basic arithmetic functions
    (add, sub, compare).

    Leap seconds are ignored, so as dayligh saving time (DST) changes.
    """

    # These are for the rfc1123/nscookie formats, which may not be localized.
    _days = 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'
    _months = 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', \
              'Oct', 'Nov', 'Dec'
    # Special "short" format for ps(1) like output.
    _shortfmt = '%H:%M', '%b%d', '%b%Y'

    _formats = { 'mysql': '%Y-%m-%d %H:%M:%S', 'ctime': '%a %b %d %H:%M:%S %Y',
                 'timestamp': '%Y%m%d%H%M%S', 'rfc1123': '%a, %d %b %Y %H:%M:%S GMT',
                 'nscookie': '%a, %d-%b-%Y %H:%M:%S GMT',
                 'iso8601': '%Y-%m-%d %H:%M:%S' }
    _nullfmt = '0000-00-00 00:00:00' # MySQL zero format. 


    def __init__(self, *args):
        """
        Initialize the DateTime object. The constructor takes either zero or
        one argument. If no argument is passed the DateTime is initialized to
        the current date and time. If an argument is given, it must be a null
        date (None or the literal string '0000-00-00 00:00:00'), a string in a
        supported format, a unix timestamp, a Date instance or an object with
        a tuple() method (like an mx.DateTime instance).
        """
        self.m_formats = self._formats.copy()
        if len(args) == 0:
            self.m_tm = list(time.localtime())
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, DateTime):
                self.m_tm = arg.m_tm[:]
            elif arg in (None, self._nullfmt):
                self.m_tm = [0]*9
            elif type(arg) in (int, long, float):
                self.m_tm = list(time.localtime(arg))
            elif type(arg) in (str, unicode):
                self.parse(args[0])
            elif hasattr(arg, 'tuple'):
                self.m_tm = map(int, arg.tuple())
            elif hasattr(arg, 'timetuple'):
                self.m_tm = list(arg.timetuple())
            else:
                raise TypeError, 'Unsupported date/time type.'
        else:
            raise TypeError, 'Expecting 0 or 1 arguments.'

    def _get_hour(self):
        return self.m_tm[3]

    def _set_hour(self, hour):
        if type(hour) != int:
            raise ValueError, 'Illegal hour: %s' % hour
        self.m_tm[3] = hour
        self.addTime(0,0,0)  # renormalize

    hour = property(_get_hour, _set_hour, None)

    def _get_minute(self):
        return self.m_tm[4]

    def _set_minute(self, minute):
        if type(minute) != int:
            raise ValueError, 'Illegal minute: %s' % minute
        self.m_tm[4] = minute
        self.addTime(0,0,0)

    minute = property(_get_minute, _set_minute, None)

    def _get_second(self):
        return self.m_tm[5]

    def _set_second(self, second):
        if type(second) != int:
            raise ValueError, 'Illegal second: %s' % second
        self.m_tm[5] = second
        self.addTime(0,0,0)

    second = property(_get_second, _set_second, None)

    def _get_time(self):
        return time.mktime(self.m_tm)

    def _set_time(self, t):
        if type(t) not in (int, long, float):
            raise TypeError, 'Illegal type for time.'
        self.m_tm = list(time.localtime(t))

    time = property(_get_time, _set_time, None)

    def addTime(self, dh, dm, ds):
        """
        Add DH hours, DM minutes and DS seconds to the current time.
        """
        if not self:
            return
        hour, minute, second = self.m_tm[3:6]
        second = hour*3600 + minute*60 + second + dh*3600 + dm*60 + ds
        dd, second = divmod(second, 86400)
        hour, second = divmod(second, 3600)
        minute, second = divmod(second, 60)
        self.m_tm[3:6] = [hour, minute, second]
        self.addDate(0, 0, dd)

    def format(self, format):
        """
        Format the DateTime according to a predefined format.
        """
        if format == 'short':
            # The 'short' format is special
            tm = list(time.localtime())
            if tm[:3] == self.m_tm[:3]:
                # Today
                return self.strftime(self._shortfmt[0])
            elif abs((tm[0]*12 + tm[1]) - (self.m_tm[0]*12 + self.m_tm[1])) < 6:
                # Within six months
                return self.strftime(self._shortfmt[1])
            else:
                # Other
                return self.strftime(self._shortfmt[2])
        elif format in ('rfc1123', 'nscookie'):
            # These formats are in GMT and must not be localized
            self.addTime(0, 0, time.timezone)
            format = self.m_formats[format]
            format = format.replace('%a', self._days[self.weekday]) \
                           .replace('%b', self._months[self.month-1])
            result =  self.strftime(format)
            self.addTime(0, 0, -time.timezone)
            return result
        else:
            return Date.format(self, format)
