# vi: ts=4 sw=4 et
#
# dbmysql.py: database access for MySQL
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: dbmysql.py,v 1.3 2003/03/18 13:57:39 geertj Exp $

import MySQLdb

import draco
from draco.database import Database, DatabaseNamespace
from draco.logger import logger

dbapi = MySQLdb

connect_params = \
{
    'DatabaseName': 'db',
    'DatabaseUser': 'user',
    'DatabasePassword': 'passwd',
    'DatabaseHost': 'host',
    'DatabasePort': 'port',
    'DatabaseSocket': 'unix_socket'
}

draco_tables = \
{
    'draco_tasks':
        """
        create table if not exists draco_tasks 
        (
            name            char(64)        not null,
            last_run        integer         not null,
            PRIMARY KEY     (name)
        ) type=MyISAM
        """,
    'draco_servers':
        """
        create table if not exists draco_servers
        (
            generation      integer         not null,
            count           integer         not null,
            restart_time    integer         not null,
            PRIMARY KEY     (generation)
        )
        """,
    'draco_sessions':
        """
        create table if not exists draco_sessions
        (
            sessionid       char(21)        not null,
            user_agent      char(32)        not null,
            create_date     datetime        not null,
            last_used       datetime        not null,       
            expires         datetime        not null,
            persistent      integer         default 0 not null,
            userid          integer         default 0 not null,
            auth_token      char(21)        default '' not null,
            PRIMARY KEY     (sessionid)
        ) type=MyISAM
        """,
    'draco_server_data':
        """
        create table if not exists draco_server_data
        (
            serverid        integer         not null,
            scope           char(64)        not null,
            name            char(64)        not null,
            value           blob            not null,
            PRIMARY KEY     (scope,name)
        ) type=MyISAM
        """,
    'draco_application_data':
        """
        create table if not exists draco_application_data
        (
            generation      integer         not null,
            scope           char(64)        not null,
            name            char(64)        not null,
            value           blob            not null,
            PRIMARY KEY     (generation,scope,name)
        ) type=MyISAM
       """,
    'draco_session_data':
        """
        create table if not exists draco_session_data 
        (
            sessionid       char(21)        not null,
            scope           char(64)        not null,
            name            char(64)        not null,
            value           blob            not null,
            PRIMARY KEY     (sessionid,scope,name)
        ) type=MyISAM
        """,
    'draco_user_data':
        """
        create table if not exists draco_user_data
        (
            userid          integer         not null,
            scope           char(64)        not null,
            name            char(64)        not null,
            value           blob            not null,
            PRIMARY KEY     (userid,scope,name)
        ) type=MyISAM
        """
}


class MySQLDatabase(Database):
    """
    Draco database access for MySQL.
    """

    def __init__(self):
        """
        Initialize the MySQL database.
        """
        self.m_connection = None

    def configure(self, config):
        """
        Configure the database object.
        """
        if self.m_connection:
            self.m_connection.close()
            self.m_connection = None
        self.m_params = {}
        for key,value in connect_params.items():
            if config.has_key(key):
                self.m_params[value] = config[key]
        self.m_persistent = config.get('DatabaseKeepConnection', 0)

    def reload(self):
        """
        Reload the database. We reconnect if there is no persistent
        connection.
        """
        if self.m_connection:
            try:
                self.m_connection.ping()
                return
            except MySQLdb.Error:
                pass
        try:
            self.m_connection = apply(MySQLdb.connect, (), self.m_params)
        except MySQLdb.Error:
            raise draco.HttpInternalServerError, \
                    'Could not connect to the database.'
        self.m_cursor = self.m_connection.cursor()

    def finalize(self):
        """
        Finalize the database for this request. The connection is closed
        if persistent database connections are not enabled.
        """
        if self.m_persistent or self.m_connection is None:
            return
        self.m_cursor.close()
        self.m_cursor = None
        self.m_connection.close()
        self.m_connection = None

    def checkTables(self):
        """
        Check if all required tables are present, creating them if necessary.
        """
        cursor = self.cursor()
        cursor.execute('show tables')
        tables = [ i[0] for i in cursor.fetchall() ]
        for ta in draco_tables:
            if ta in tables:
                continue
            logger.info('Creating system table %s' % ta)
            cursor.execute(draco_tables[ta])

    def createNamespace(self, table, key, value, scope):
        """
        Return a MySQL namespace on top of table `table', forcing
        `key' to `value' and with scope `scope'.
        """
        return MySQLNamespace(self, table, key, value, scope)

    def connection(self):
        """
        Return the database connection.
        """
        return self.m_connection

    def cursor(self):
        """
        Return the global database cursor.
        """
        return self.m_cursor

    def lock(self):
        """
        Lock the database. This is used for implementing application locks.
        """
        cursor = self.cursor()
        cursor.execute('lock tables draco_application_data write')

    def unlock(self):
        """
        Unlock the database.
        """
        cursor = self.cursor()
        cursor.execute('unlock tables')


class MySQLNamespace(DatabaseNamespace):
    """
    A MySQL optimized namespace.
    """

    def available(self):
        return not not self.value

    def __len__(self):
        if not self.available():
            raise draco.HttpInternalServerError, 'Namespace not available.'
        cursor = self.database.cursor()
        cursor.execute("""
                select count(*) from %s
                where scope = %%s and %s = %%s
                """ % (self.table, self.key),
                      (self.scope, self.value))
        return cursor.fetchone()[0]

    def __getitem__(self, name):
        if not self.available():
            raise draco.HttpInternalServerError, 'Namespace not available.'
        cursor = self.database.cursor()
        cursor.execute("""
                select value from %s
                where scope = %%s and name = %%s and %s = %%s
                """ % (self.table, self.key),
                      (self.scope, name, self.value))
        row = cursor.fetchone()
        if row is None:
            raise KeyError, name
        return self.deserialize(row[0])

    def __setitem__(self, name, value):
        if not self.available():
            raise draco.HttpInternalServerError, 'Namespace not available.'
        data = self.serialize(value)
        cursor = self.database.cursor()
        cursor.execute("""
                replace into %s
                (%s, scope, name, value) values (%%s, %%s, %%s, %%s)
                """ % (self.table, self.key),
                      (self.value, self.scope, name, data))

    def __delitem__(self, name):
        if not self.available():
            raise draco.HttpInternalServerError, 'Namespace not available.'
        cursor = self.database.cursor()
        cursor.execute("""
                delete from %s
                where %s = %%s and scope = %%s and name = %%s
                """ % (self.table, self.key),
                      (self.value, self.scope, name))
        if not cursor.rowcount:
            raise KeyError, name

    def __contains__(self, name):
        if not self.available():
            raise draco.HttpInternalServerError, 'Namespace not available.'
        cursor = self.database.cursor()
        cursor.execute("""
                select count(*) from %s
                where %s = %%s and scope = %%s and name = %%s
                """ % (self.table, self.key),
                      (self.value, self.scope, name))
        count = cursor.fetchone()[0]
        return count != 0

    def items(self):
        if not self.available():
            raise draco.HttpInternalServerError, 'Namespace not available.'
        cursor = self.database.cursor()
        cursor.execute("""
                select name, value from %s
                where %s = %%s and scope = %%s
                """ % (self.table, self.key),
                      (self.value, self.scope))
        data = cursor.fetchall()
        result = [ (name, self.deserialize(value)) for name,value in data ]
        return result

    def clear(self):
        if not self.available():
            raise draco.HttpInternalServerError, 'Namespace not available.'
        cursor = self.database.cursor()
        cursor.execute("""
                delete from %s
                where %s = %%s and scope = %%s
                """ % (self.table, self.key),
                      (self.value, self.scope))
