# vi: ts=4 sw=4 et
#
# document.py: document parser
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: document.py,v 1.44 2003/03/26 09:36:09 geertj Exp $

import os
import re
import types
import codecs

import draco
from draco.util import reindent


class DocumentError(draco.HttpInternalServerError):
    """
    A document related error occurred.
    """

class ParseError(DocumentError):
    """
    An error occurred parsing a document.
    """


class Tag:
    """
    A Tag as passed to tag rewriter functions. The tag can be modified
    in two simple ways. First, the tag attributes can be modified arbitrarily
    and second a literal text can be append after the opening tag.
    """

    def __init__(self, name, attrs={}, endstyle='>'):
        """
        Create a new tag name with attributes attrs.
        """
        self.name = name
        self.attrs = attrs
        self.endstyle = endstyle
        self.buffer = ''

    def appendText(self, text):
        """
        Append some literal text after the opening tag.
        """
        self.buffer += text

    def __str__(self):
        """
        Output the tag as a string.
        """
        result = '<' + self.name
        for name, value in self.attrs.items():
            result += ' %s' % name
            if value is not None:
                if value.find('"') >= 0:
                    result += "='%s'" % value
                else:
                    result += '="%s"' % value
        result += self.endstyle + self.buffer
        return result


class FileOpener:
    """
    An opener class to open documents from the filesystem. Documents are 
    opened relative to $UserRoot if their name begins with a '/', or 
    otherwise relative to the main document directory. Unicode files starting
    with a BOM (byte order mark) are supported.
    """

    def __init__(self, rootdir, curdir):
        """
        Initialize the file opener. ROOTDIR is the root directory, relative to
        which absolute includes are resolved. CURDIR is the directory to which
        relative includes are resolved.
        """
        self.m_rootdir = rootdir
        if not self.m_rootdir.endswith('/'):
            self.m_rootdir += '/'
        self.m_curdir = curdir
        if not self.m_curdir.endswith('/'):
            self.m_curdir += '/'

    def _expand(self, name):
        """
        Expand NAME to a full pathname.
        """
        if name[0] == '/':
            fname = self.m_rootdir + name
        else:
            fname = self.m_curdir + name
        return fname

    def exists(self, fname):
        """
        Return nonzero if FNAME exists.
        """
        fname = self._expand(fname)
        try:
            os.stat(fname)
        except OSError:
            return 0
        return 1

    def open(self, fname):
        """
        Open a document and return the file object. We try to detect if the
        file is unicode by looking for a BOM (Byte Order Mark.
        """
        fname = self._expand(fname)
        file = open(fname, 'r')
        head = file.read(3)
        if head[:2] == '\xff\xfe':
            enc = 'UTF-16LE'
        elif head[:2] == '\xfe\xff':
            enc = 'UTF-16BE'
        elif head == '\xef\xbb\xbf':
            enc = 'UTF-8'
        else:
            enc = None
        if enc:
            file.close()
            file = codecs.open(fname, 'r', enc)
            file.read(1)  # Skip BOM
        else:
            file.seek(0)
        return file


class Parser:
    """
    A parser for draco documents. 
    """

    def __init__(self, document, opener=None, environment=None, rewriters=None):
        """
        Initialize the parser.  DOCUMENT is the main document to parse. 
        If OPENER is specified, DOCUMENT must be the name of a document
        that can be opened with it. Otherwise, DOCUMENT must be a string
        that is parsed directly. No includes are handled in this case.
        ENVIRONMENT must be a Environment instance that has eval() and
        run() methods. It is used to run embedded code and expressions.
        If it is not specified, no code or expressions are evaluated. 
        REWRITE must be a Rewriter instance. If it is not specified, no
        tags are rewritten. 
        """
        self.m_document = document
        self.m_opener = opener
        self.m_environment = environment
        self.m_rewriters = rewriters or []

    def parse(self):
        """
        Parse the document and return the result in a string.
        """
        if self.m_opener:
            document = self.m_opener.open(self.m_document).read()
            filename = self.m_opener._expand(self.m_document)
        else:
            document = self.m_document
            filename = '<string>'
        pass1 = self.parseEmbedded(document, filename)
        pass2 = self.parseTags(pass1)
        return pass2

    def parseEmbedded(self, data, filename, include_path=[]):
        """
        Parse a text string for <% %> embedded codes.
        """
        result = data[0:0]  # Use same type as data
        p0 = 0
        linenr = 0
        while 1:
            p1 = data.find('<%', p0)
            if p1 < 0:
                result += data[p0:]
                break
            result += data[p0:p1]
            linenr += data[p0:p1].count('\n')
            p1 += 2
            p2 = data.find('%>', p1)
            if p2 < 0:
                raise ParseError, 'Unmatched <% tag.'
            if data[p1] == '@':
                result += self.parseDirective(data[p1+1:p2], filename,
                                linenr, include_path)
            elif data[p1] == '=':
                res = self.parseExpression(data[p1+1:p2], filename, linenr)
                result += res
            else:
                res = self.parseCode(data[p1:p2], filename, linenr)
                result += res
            p0 = p2+2
            linenr += data[p1:p2].count('\n')
        return result

    re_attr = re.compile("""
        \s*
        ([a-z_:][-a-z0-9._:]*)
        (?:
          \s*=\s*
          (
            [-a-z0-9._:]+  |
            "[^"]*"      |
            '[^']*'
          )
        )?""", re.IGNORECASE|re.VERBOSE)

    def parseAttributes(self, data):
        """
        Parse an HTML attribute string and return the results in
        a dictionary.
        """
        attrs = {}
        result = self.re_attr.findall(data)
        for name, val in result:
            if val:
                if val[0] in '\'"':
                    val = val[1:-1]
                val = str(val)  # Get rid of evt. unicode
            else:
                val = None
            attrs[str(name)] = val
        return attrs

    def parseEndStyle(self, data):
        """
        Parse the tag end style (important with xhtml documents).
        """
        if data[-2:] == ' /':
            return ' />'
        elif data[-1:] == '/':
            return '/>'
        else:
            return '>'

    def parseDirective(self, data, filename, linenr, include_path):
        """
        Parse an embedded directive. An embedded directive looks like this:
        <%@ include file = "header.inc" %>
        """
        p0 = 0
        try:
            while data[p0].isspace():
                p0 += 1
            p1 = p0
            while data[p1].isalpha():
                p1 += 1
        except IndexError:
            raise ParseError, 'Parse error in embedded directive.'
        directive = data[p0:p1]
        attrs = self.parseAttributes(data[p1:])
        if directive == 'include':
            if not self.m_opener:
                return ''
            if attrs.has_key('file'):
                name = attrs['file']
            elif attrs.has_key('expr'):
                name = self.parseExpression(attrs['expr'], filename, linenr)
            else:
                raise DocumentError, 'Include doesn\'t specify file.'
            if not name:
                return ''
            if name in include_path:
                raise DocumentError, 'Include loop detected.'
            include_path = include_path + [name]  # create a copy
            # Parse this file too
            try:
                data = self.m_opener.open(name).read()
            except IOError:
                raise DocumentError, 'Could not open include file: %s.' % name
            fname = self.m_opener._expand(name)
            return self.parseEmbedded(data, fname, include_path)
        else:
            raise ParseError, 'Unknown directive: %s.' % directive

    def parseExpression(self, data, filename, linenr):
        """
        Run an expression. An embedded expression looks like <%= expr %>.
        We prepend LINENR newlines before the code so that in case of an
        exception a traceback will the corrent line number. This is very
        ugly of course but provides an invaluable debugging aid. 
        """
        if not self.m_environment:
            return ''
        code = '\n' * linenr + reindent(data, strip=0) + '\n'
        result = self.m_environment.eval(code, filename)
        if type(result) not in (str, unicode):
            result = str(result)
        return result

    def parseCode(self, data, filename, linenr):
        """
        Run a clode block. An embedded code block looks like <% code %>.
        """
        if not self.m_environment:
            return ''
        code = '\n' * linenr + reindent(data, strip=0) + '\n'
        result = self.m_environment.run(code, filename)
        return result

    def getRewriterDict(self):
        """
        Return a dictionary containing lists with rewriter methods.
        """
        rwdict = {}
        for rw in self.m_rewriters:
            for name in dir(rw):
                attr = getattr(rw, name)
                if callable(attr) and not name.startswith('__'):
                    try:
                        rwdict[name].append(attr)
                    except KeyError:
                        rwdict[name] = [attr]
        return rwdict

    def parseTags(self, data):
        """
        Find all html tags and call a rewriter for them, if installed.
        """
        rwdict = self.getRewriterDict()
        if not rwdict:
            return data
        p0 = 0
        result = data[0:0]  # Same type as data (unicode or str)
        pattern = '<(' + '|'.join(rwdict.keys()) + ')[\s>]'
        re_tags = re.compile(pattern, re.IGNORECASE)
        while 1:
            match = re_tags.search(data, p0)
            if not match:
                result += data[p0:]
                break
            result += data[p0:match.start(0)]
            p0 = match.end(1)
            p1 = data.find('>', p0)
            if p1 < 0:
                raise ParseError, 'Premature EOF encountered.'
            attrs = self.parseAttributes(data[p0:p1])
            endstyle = self.parseEndStyle(data[p0:p1])
            tag = Tag(match.group(1), attrs, endstyle)
            for rewrite in rwdict[tag.name]:
                rewrite(tag)
            result += str(tag)
            p0 = p1+1
        return result
