# vi: ts=4 sw=4 et
#
# draco.py: the mod_python handler
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: dracohandler.py,v 1.67 2003/03/26 09:36:25 geertj Exp $

import os
import os.path
import sys
import hotshot
import traceback

import draco
import draco.encode

from mod_python import apache
from mod_python import util as mputil

config_setup = 0
child_setup = 0
app_setup = 0

config_error_message = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN"
        "http://www.w3.org/TR/REC-html40/strict.dtd">
<HTML>
  <HEAD>
    <TITLE>Incomplete Configuration</TITLE>
  </HEAD>
  <BODY>
    <H1>Incomplete Configuration</H1>
    <P> The required PythonOption <em>DocumentRoot</em> is not given in the
    Apache configuration file. Please add the following line to the apache
    configuration of this server:
    <PRE>    PythonOption DocumentRoot /path/to/documentroot</PRE>
    with <TT>/path/to/documentroot</TT> pointing to your document root.
    <HR> <ADDRESS>Draco/%s at %s port %d</ADDRESS>
  </BODY>
</HTML>
"""
client_error_message = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN"
        "http://www.w3.org/TR/REC-html40/strict.dtd">
<HTML>
  <HEAD>
    <TITLE>Client Error</TITLE>
  </HEAD>
  <BODY>
    <H1>Client Error</H1>
    <PRE>Exception: %s
Message: %s</PRE>
    <HR> <ADDRESS>Draco/%s at %s port %d</ADDRESS>
  </BODY>
</HTML>
"""
server_error_message = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN"
        "http://www.w3.org/TR/REC-html40/strict.dtd">
<HTML>
  <HEAD>
    <TITLE>Server Error</TITLE>
  </HEAD>
  <BODY>
    <H1>Server Error</H1>
    <PRE>Exception: %s
Message: %s</PRE>
    <HR> <ADDRESS>Draco/%s at %s port %d</ADDRESS>
  </BODY>
</HTML>
"""
other_error_message = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN"
        "http://www.w3.org/TR/REC-html40/strict.dtd">
<HTML>
  <HEAD>
    <TITLE>Uncaught Exception</TITLE>
  </HEAD>
  <BODY>
    <H1>Uncaught Exception</H1>
    <PRE>Exception: %s
Value: %s</PRE>
    <H1>A traceback is provided below:</H1>
    <PRE>%s</PRE>
    <HR> <ADDRESS>Draco/%s at %s port %d</ADDRESS>
  </BODY>
</HTML>
"""


def format_trace(trace):
    """
    Format a traceback.
    """
    trace = traceback.format_tb(trace)
    trace.insert(0, 'Traceback (most recent call last):\n')
    trace = ''.join(trace)
    return trace


def log_exception():
    """
    Write a trackback of the current exception to the system log.
    """
    exc, value, trace = sys.exc_info()
    try:
        name = exc.__name__
    except AttributeError:
        name = str(exc)
    value = str(value)
    trace = format_trace(trace)
    logger.error("""** An uncaught exception occurred **
            Exception: %s
            Value: %s\n
            %s
            """, (name, value, trace))


def setupConfig(mpreq):
    """
    Setup up the configuration and the logging subsystem. This is done once
    for each child.
    """
    options = mpreq.get_options()
    if not options.has_key('DocumentRoot'):
        mpreq.content_type = 'text/html'
        mpreq.send_http_header()
        try:
            mpreq.write(config_error_message % (draco.cfg_version, \
                    mpreq.server.server_hostname, mpreq.server.port))
        except IOError:
            pass
        raise apache.SERVER_RETURN, apache.OK
    global logger
    logfile = options['DocumentRoot'] + '/error.log'
    from draco import logger as mlogger
    logger = mlogger.logger = mlogger.Logger(logfile)
    global config
    cfgfiles = draco.cfg_datadir + '/draco.ini', \
            options['DocumentRoot'] + '/draco.ini'
    from draco import config as mconfig
    config = mconfig.config = mconfig.Config(cfgfiles)


def reloadConfig(mpreq):
    """
    Reload the configuration subsystem.
    """
    config.reload(mpreq.get_options())


def setupChild(mpreq):
    """
    Set up the web server child. This instantiates all global draco objects
    and performs other one-time initializations.
    """
    global loader, database, dbapi
    global client, server, cookies, request, response, application, session, user
    global eventmanager, handlermanager, rewritemanager, scheduler, imagemanager
    sysroot = draco.cfg_datadir + '/lib'
    docroot = config['DocumentRoot']
    from draco import loader as mloader
    loader = mloader.loader = mloader.Loader()
    from draco import database as mdatabase
    factory = mdatabase.DatabaseFactory()
    factory.configure(config)
    database = mdatabase.database = factory.createDatabase()
    dbapi = mdatabase.dbapi = factory.createDbapi()
    from draco import client as mclient
    client = mclient.client = mclient.Client()
    from draco import server as mserver
    server = mserver.server = mserver.Server()
    from draco import cookie as mcookie
    cookies = mcookie.cookies = mcookie.Cookies()
    from draco import request as mrequest
    request = mrequest.request = mrequest.Request()
    from draco import response as mresponse
    response = mresponse.response = mresponse.Response()
    from draco import session as msession
    session = msession.session = msession.Session()
    from draco import application as mapplication
    application = mapplication.application = mapplication.Application()
    from draco import user as muser
    user = muser.user = muser.User()
    from draco import event as mevent
    evfiles = sysroot + '/__events__.py', docroot + '/__events__.py'
    eventmanager = mevent.eventmanager = mevent.EventManager(evfiles)
    from draco import rewrite as mrewrite
    rwfiles = sysroot + '/__rewrite__.py', docroot + '/__rewrite__.py'
    rewritemanager = mrewrite.rewritemanager = mrewrite.RewriteManager(rwfiles)
    from draco import task as mtask
    tafiles = sysroot + '/__tasks__.py', docroot + '/__tasks__.py'
    scheduler = mtask.scheduler = mtask.Scheduler(tafiles)
    from draco import image as mimage
    imagemanager = mimage.imagemanager = mimage.ImageManager(docroot)
    from draco import handler as mhandler
    handlermanager = mhandler.handlermanager = mhandler.HandlerManager()
    global FileOpener, Parser, DracoEnvironment
    from draco.document import FileOpener, Parser
    from draco.environment import DracoEnvironment
    sys.path.insert(1, os.path.normpath(config['DocumentRoot'] + '/__lib__'))
    mpreq.server.register_cleanup(mpreq, childExit, None)


def reloadChild(mpreq):
    """
    Reload all global draco objects.
    """
    configure = config.modified() or not child_setup
    if configure:
        loader.configure(config)
        database.configure(config)
    loader.reload()
    database.reload()
    if configure:
        database.checkTables()
        client.configure(config)
        server.configure(config)
        cookies.configure(config)
        request.configure(config)
        response.configure(config)
        application.configure(config)
        session.configure(config)
        user.configure(config)
    client.reload(mpreq)
    server.reload(mpreq)
    cookies.reload()
    request.reload(mpreq)
    response.reload(mpreq)
    application.reload()
    session.reload()
    user.reload()
    if configure:
        eventmanager.configure(config)
        rewritemanager.configure(config)
        scheduler.configure(config)
        handlermanager.configure(config)
        imagemanager.configure(config)
    eventmanager.reload()
    rewritemanager.reload()
    scheduler.reload()
    handlermanager.reload()
    imagemanager.reload()


def setupApplication():
    """
    Per application initialization. We return 1 if we are the first
    webserver in this generation.
    """
    if not database or server.generation() is None:
        logger.info('Application variables and events are not available.')
        return 0
    cursor = database.cursor()
    cursor.execute("""
            select count, restart_time
            from draco_servers
            where generation = %s
            """, (server.generation(),))
    row = cursor.fetchone()
    if row and row[1] != server.restart():
        cursor.execute("""
                delete from draco_servers
                where generation = %s and restart_time = %s
                """, (server.generation(), row[1]))
        database.commit()
        row = None
    if not row:
        try:
            cursor.execute("""
                    insert into draco_servers
                    (generation, count, restart_time)
                    values (%s, 1, %s)
                    """, (server.generation(), server.restart()))
            database.commit()
            cursor.execute('delete from draco_application_data')
            database.commit()
            return 1
        except dbapi.IntegrityError:
            # Another process inserted a row for this generation
            database.rollback()
    cursor.execute("""
            update draco_servers
            set count = count + 1
            where generation = %s
            """, (server.generation(),))
    database.commit()
    return 0


def childExit(data):
    """
    Called when a web server child exits. There is no client at the moment
    so we have to reconnect to the database.
    """
    if server.generation() is None:
        return
    try:
        database.reload()
        try:
            cursor = database.cursor()
            cursor.execute("""
                    select count
                    from draco_servers
                    where generation = %s
                    for update
                    """, (server.generation(),))
            count = cursor.fetchone()[0]
            cursor.execute("""
                    update draco_servers
                    set count = count - 1
                    where generation = %s
                    """, (server.generation(),))
            database.commit()
            eventmanager.raiseEvent('childEnd')
            if count == 1:
                eventmanager.raiseEvent('applicationEnd')
        finally:
            database.finalize()
    except:
        log_exception()


def reapChildren():
    """
    Reap all available child processes.
    """
    try:
        pid = 1
        while pid:
            pid, status = os.waitpid(-1, os.WNOHANG)
    except OSError:
        pass


def accesshandler(mpreq):
    """
    Apache's "Access" handler. We use it to hide our special files from the 
    web space.
    """
    file = mpreq.filename
    if not file or file == '/':
        return apache.DECLINED
    # Check for forbidden extensions
    pos = file.rfind('.')
    if pos > 0:
        ext = file[pos:]
        if ext in ('.ini', '.inc', '.log', '.py', '.pyc', '.pyo', '.pyw'):
            return apache.HTTP_FORBIDDEN
    # Check for a '__xxx' path element which is also forbidden.
    path = file.split('/')[1:]
    for i in path:
        if i[:2] == '__':
            return apache.HTTP_FORBIDDEN
    return apache.DECLINED


def handler(mpreq):
    """
    The mod_python handler that calls or profiles the main _handler.
    """
    global config_setup
    if not config_setup:
        setupConfig(mpreq)
        config_setup = 1
    reloadConfig(mpreq)

    # Clean up child processes that the user might have started.
    reapChildren()

    # Do we handle this request?
    docroot = config['DocumentRoot']
    if not mpreq.filename.startswith(docroot):
        return apache.DECLINED
    ext = '.' + config.get('Extension', 'dsp')
    if not mpreq.filename[-4:] == ext:
        return apache.DECLINED
    if not mpreq.method in ('GET', 'HEAD', 'POST'):
        return apache.DECLINED

    # Call or profile the _handler function, depending on user config
    do_profile = config.get('Profile') and config.get('ProfileOutput')
    if do_profile:
        logfile = config.get('ProfileOutput')
        profiler = hotshot.Profile(logfile)
        ret = profiler.runcall(_handler, mpreq)
        profiler.close()
    else:
        ret = _handler(mpreq)
    return ret


def _handler(mpreq):
    """
    The actual draco handler.
    """
    debug = config.get('Debug', 1)
    hostname = mpreq.server.server_hostname
    http_port = mpreq.connection.local_addr[1]

    try:
        # Setup child and application
        newapp = 0
        newchild = 0

        global child_setup
        if not child_setup:
            setupChild(mpreq)
            newchild = child_setup = 1

        try:
            reloadChild(mpreq)

            global app_setup
            if not app_setup:
                newapp = setupApplication()
                app_setup = 1

            # Raise events
            if newapp:
                eventmanager.raiseEvent('applicationStart')
            if newchild:
                eventmanager.raiseEvent('childStart')
            elif config.modified():
                eventmanager.raiseEvent('configChanged')
            if session and session.isnew():
                eventmanager.raiseEvent('sessionStart')
            eventmanager.raiseEvent('requestStart')

            # Run scheduled tasks
            scheduler.runTasks()

            # Load handler
            hclass = handlermanager.loadHandler(request.dir())
            if hclass:
                handler = hclass()
                page = request.filename()[:-4]
                if hasattr(handler, '__authenticate__'):
                    method = getattr(handler, '__authenticate__')
                    auth = method(page)
                    if auth != handler.AUTH_OK:
                        raise draco.HttpForbidden, 'Access not allowed.'
                # Require the handler to be implemented in this class for
                # security reasons. This way it is impossible to call base
                # methods like __setitem__ etc.
                if hclass.__dict__.has_key(page):
                    method = getattr(handler, page)
                    method(request.pathinfo(), request.args())
                elif hasattr(handler, '__default__'):
                    method = getattr(handler, '__default__')
                    method(request.pathinfo(), request.args())
                else:
                    handler = None
            else:
                handler = None

            # Load template
            docroot = config['DocumentRoot']
            opener = FileOpener(docroot, request.dir())
            document = request.filename()
            template = opener.exists(document)
            if not template and handler is None:
                raise draco.HttpNotFound, 'Document not found.'
            if template:
                environment = DracoEnvironment()
                if handler:
                    environment.update(handler)
                response.addHeader('Cache-Control', 'no-cache')
                rewriters = rewritemanager.rewriters()
                parser = Parser(document, opener, environment, rewriters)
                output = parser.parse()
                response.setBuffering(1)
                response.write(output)
            response.flush()

            # We are done
            eventmanager.raiseEvent('requestEnd')
            return apache.OK

        finally:
            database.finalize()

    except draco.HttpClientError, exc:
        name = exc.__class__.__name__
        value = str(exc)
        logger.info("""** A client error occurred **
                Exception: %s
                Value: %s
                """, (name, value))
        if debug:
            mpreq.content_type = 'text/html'
            mpreq.send_http_header()
            try:
                mpreq.write(client_error_message % \
                        (name.encode('html'), value.encode('html'), \
                         draco.cfg_version, hostname, http_port))
            except IOError:
                pass
            return apache.OK
        return exc.code

    except draco.HttpServerError, exc:
        name = exc.__class__.__name__
        value = str(exc) + '\n'
        if hasattr(exc, 'filename'):
            value += 'File: %s\n' % exc.filename
        if hasattr(exc, 'lineno'):
            value += 'Line: %d\n' % exc.lineno
        if hasattr(exc, 'traceback'):
            value += '\n%s\n' % exc.traceback
        logger.warn("""** A server error occurred **
                Exception: %s
                Value: %s
                """, (name, value))
        if debug:
            mpreq.content_type = 'text/html'
            mpreq.send_http_header()
            try:
                mpreq.write(server_error_message % \
                        (name.encode('html'), value.encode('html'), \
                         draco.cfg_version, hostname, http_port))
            except IOError:
                pass
            return apache.OK
        return exc.code

    except draco.HttpResponse, err:
        if err.code == 200:
            return apache.OK
        else:
            return err.code

    except:
        exc, value, trace = sys.exc_info()
        try:
            name = exc.__name__
        except AttributeError:
            name = str(exc)
        value = str(value)
        trace = format_trace(trace)
        logger.error("""** An uncaught exception occurred **
                Exception: %s
                Value: %s\n
                %s
                """, (name, value, trace))
        if debug:
            mpreq.content_type = 'text/html'
            mpreq.send_http_header()
            try:
                mpreq.write(other_error_message % (name.encode('html'), \
                         value.encode('html'), trace.encode('html'), \
                         draco.cfg_version, hostname, http_port))
            except IOError:
                pass
            return apache.OK
        return apache.HTTP_INTERNAL_SERVER_ERROR
