# vi: ts=4 sw=4 et
#
# encode.py: Various web related encodings/decodings.
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: encode.py,v 1.7 2003/03/10 16:38:37 geertj Exp $
# vi: ts=8 sts=4 sw=4

from codecs import register, Codec
from codecs import StreamReader, StreamWriter


safe_chars = ('ABCDEFGHIJKLMNOPQRSTUVWXYZ'
              'abcdefghijklmnopqrstuvwxyz'
              '0123456789' '_.-')


def quote_hex(s, safe=''):
    """
    Replace potentially unsafe characters in 's' with their %XX hexadecimal
    counterparts. You can pass additional safe characters in `safe'.
    """
    res = list(s)
    safe += safe_chars
    for i in range(len(s)):
        c = res[i]
        if c not in safe:
            res[i] = '%%%02X' % ord(c)
    return ''.join(res)


def unquote_hex(s):
    """
    Change %XX occurences in `s' with their character value. 
    Does the opposite of quote_url().
    """
    list = s.split('%')
    res = [list[0]]
    del list[0]
    for i in list:
        if len(i) >= 2:
            try:
                c = chr(int(i[:2], 16))
                res.append(c + i[2:])
            except:
                res.append('%' + i)
        else:
            res.append('%' + i)
    return ''.join(res)
        

# (Stream)Codecs for 'url', 'form' 'html' and 'attr' encodings.

class UrlCodec(Codec):
    """
    URL encoding codec. URL encoding encodes non-safe characters in %xx
    escape sequences.
    """

    def encode(input, errors=None):
        return quote_hex(input, '/'), len(input)

    encode = staticmethod(encode)

    def decode(input, errors=None):
        return unquote_hex(input), len(input)

    decode = staticmethod(decode)


class UrlStreamReader(UrlCodec, StreamReader):
    pass


class UrlStreamWriter(UrlCodec, StreamWriter):
    pass


class FormCodec(Codec):
    """
    Form encoding. This is similar to url encoding but ' ' is encoded
    as '+' and '/' is considered unsafe.
    """

    def encode(input, errors=None):
        ilen = len(input)
        input = quote_hex(input, ' ')
        return input.replace(' ', '+'), ilen
        
    encode = staticmethod(encode)

    def decode(input, errors=None):
        ilen = len(input)
        input = input.replace('+', ' ')
        return unquote_hex(input), ilen

    decode = staticmethod(decode)


class FormStreamReader(FormCodec, StreamReader):
    pass


class FormStreamWriter(FormCodec, StreamWriter):
    pass


class HtmlCodec(Codec):
    """
    Html encoding, replace unsafe html characters with their unicode
    entities.
    """

    def encode(input, errors=None):
        ilen = len(input)
        input = input.replace('&', '&amp;').replace('<', '&lt;'). \
                      replace('>', '&gt;')
        return input, ilen

    encode = staticmethod(encode)

    def decode(input, errors=None):
        ilen = len(input)
        input = input.replace('&gt;', '>').replace('&lt;', '<'). \
                      replace('&amp;', '&')
        return input, ilen

    decode = staticmethod(decode)
        

class HtmlStreamReader(HtmlCodec, StreamReader):
    pass


class HtmlStreamWriter(HtmlCodec, StreamWriter):
    pass


class AttrCodec(Codec):
    """
    Html attribute encoding, replace " by &quot;.
    """

    def encode(input, errors=None):
        return input.replace('"', '&quot;'), len(input)

    encode = staticmethod(encode)

    def decode(input, errors=None):
        return input.replace('&quot;', '"'), len(input)

    decode = staticmethod(decode)


class AttrStreamReader(AttrCodec, StreamReader):
    pass


class AttrStreamWriter(AttrCodec, StreamWriter):
    pass


# Register the codecs

encodings = \
{ 
    'url': (UrlCodec.encode, UrlCodec.decode, UrlStreamReader, UrlStreamWriter),
    'form': (FormCodec.encode, FormCodec.decode, FormStreamReader, FormStreamWriter),
    'html': (HtmlCodec.encode, HtmlCodec.decode, HtmlStreamReader, HtmlStreamWriter),
    'attr': (AttrCodec.encode, AttrCodec.decode, AttrStreamReader, AttrStreamWriter)
}
    
def get_codec(name):
    try:
        return encodings[name]
    except:
        pass

register(get_codec)
