# vi: ts=4 sw=4 et
#
# environment.py: Different environments for running code.
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: environment.py,v 1.35 2003/03/24 08:04:34 geertj Exp $

import sys
import os
import os.path

from StringIO import StringIO
from rexec import RExec

import draco
from draco import util
from draco.namespace import DictNamespace


class CodeError(draco.HttpInternalServerError):
    """
    An error occurred when executing Python code.
    """


class Environment:
    """
    Abstract base class for all environments.
    """

    def startCapture(self):
        """
        Start capturing output written to stdout and stderr.
        """
        self.m_fds = sys.stdin, sys.stdout, sys.stderr
        self.m_buf = StringIO()
        sys.stdin = None
        sys.stdout = self.m_buf
        sys.stderr = self.m_buf

    def endCapture(self):
        """
        End the caputure.
        """
        sys.stdin, sys.stdout, sys.stderr = self.m_fds

    def output(self):
        """
        Retuns the captured output.
        """
        return self.m_buf.getvalue()

    def eval(self, expression, filename):
        """
        Evaluate the expression EXPR in the environment and return the result.
        """
        raise NotImplementedError, 'eval() must be implemented in a subclass.'

    def run(self, code, filename):
        """
        Run the code block CODE in the environment and return the output.
        """
        raise NotImplementedError, 'run() must be implemented in a subclass.'


class DefaultEnvironment(Environment, DictNamespace):
    """
    The default, unrestricted environment.
    """

    def __init__(self):
        DictNamespace.__init__(self)
        self['__name__'] = '__main__'

    def processCode(self, code):
        pass

    def eval(self, expr, filename):
        try:
            code = compile(expr, filename, 'eval')
            self.processCode(code)
            globals = self.copy()
            result = eval(code, globals)
            self.update(globals)
        except SyntaxError, err:
            error = CodeError('Syntax error in expression.')
            error.filename = filename
            error.lineno = err.lineno
            raise error
        except:
            error = CodeError('Uncaught exception in expression.')
            error.filename = filename
            error.traceback = util.getBacktrace()
            raise error
        return result

    def run(self, code, filename):
        self.startCapture()
        try:
            try:
                code = compile(code, filename, 'exec')
                self.processCode(code)
                globals = self.copy()
                eval(code, globals)
                self.update(globals)
            except SyntaxError, err:
                error = CodeError('Syntax error in code block.')
                error.filename = filename
                error.lineno = err.lineno
                raise error
            except:
                error = CodeError('Uncaught exception in code block.')
                error.filename = filename
                error.traceback = util.getBacktrace()
                raise error
        finally:
            self.endCapture()
        return self.output()


class DracoEnvironment(DefaultEnvironment):
    """
    DracoEnvironment provides access to all Draco objects and fills
    in default values for uninitialized variables.
    """

    def __init__(self):
        DefaultEnvironment.__init__(self)
        self['draco'] = draco
        from draco.config import config
        self['config'] = config
        from draco.database import database
        self['database'] = database
        from draco.client import client
        self['client'] = client
        from draco.server import server
        self['server'] = server
        from draco.cookie import cookies
        self['cookies'] = cookies
        from draco.request import request
        self['request'] = request
        from draco.response import response
        self['response'] = response
        from draco.application import application
        self['application'] = application
        from draco.session import session
        self['session'] = session
        from draco.user import user
        self['user'] = user

    def processCode(self, code):
        """
        This function sets default values for uninitialized variables.
        """
        for va in code.co_names:
            if va not in self and '.' not in va and va not in __builtins__:
                self[va] = ''
        for co in code.co_consts:
            if type(co) == type(code):
                self.processCode(co)


class RestrictedEnvironment(RExec, Environment, DictNamespace):
    """
    RestrictedEnvironment provides a restricted execution environment using
    pythons RExec functionality.
    """

    ok_posix_names = ()
    # Stdin/out/err are redirected so they are safe.
    ok_sys_names = RExec.ok_sys_names + ('stdin', 'stdout', 'stderr')

    def __init__(self):
        """
        Construct a new restricted environment. 
        """
        RExec.__init__(self)
        DictNamespace.__init__(self)

    def r_open(self, file, mode='r', buf=-1):
        raise IOError, 'Cannot open files'

    def __setitem__(self, name, value):
        DictNamespace.__setitem__(self, name, value)
        module = self.add_module('__main__')
        setattr(module, name, value)
        
    def __delitem__(self, name):
        DictNamespace.__delitem__(self, name)
        module = self.add_module('__main__')
        delattr(module, name)

    def eval(self, expr, filename):
        try:
            result = self.r_eval(expr)
        except SyntaxError, err:
            error = CodeError('Syntax error in expression.')
            error.filename = filename
            error.lineno = err.lineno
            raise error
        except:
            error = CodeError('Uncaught exception in expression.')
            error.filename = filename
            error.traceback = util.getBacktrace()
            raise error
        return result

    def run(self, code, filename):
        self.startCapture()
        try:
            try:
                self.r_exec(code)
            except SyntaxError, err:
                error = CodeError('Syntax error in code block.')
                error.filename = filename
                error.lineno = err.lineno
                raise error
            except:
                error = CodeError('Uncaught exception in code block.')
                error.filename = filename
                error.traceback = util.getBacktrace()
                raise error
        finally:
            self.endCapture()
        return self.output()
