# vi: ts=4 sw=4 et
#
# event.py: Contains Event and EventManager.
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: event.py,v 1.21 2003/03/10 16:38:37 geertj Exp $

import types

from draco.logger import logger
from draco.loader import loader
from draco.filewatch import FileWatch
from draco.util import hiercmp, getBacktrace


class Event:
    """
    Base class for all events. Events can be implemented by subclassing this
    class and providing one or more of the following methods:

    applicationStart(): 

        Called when the web server starts or restarts. Usefull for setting
        application wide variables.

    applicationEnd(): 

        Called when the web server exits or is being restarted.

    childStart():

        Called after a new web server child has been created. Usefull for 
        setting persistent python variables.

    childEnd():

        Called just before a web server child exits.

    sessionStart(): 

        Called when a new session has been created. The session is available
        using "from session import session". Usefull for setting session wide
        variables.

    sessionEnd():

        Called when a session is being expired. The session is available
        using "from session import session". This event is currently not
        supported.

    requestStart():

        Called for each request, just before a handler or embedded code is run.
        Usefull for initializing per-request python variables.

    requestEnd():

        Called at the end of each request, just after the content has been sent
        to the client.
    """


class EventManager:
    """
    Global event manager. It loads the event handlers from the files passed
    to the constructor and keeps them up to date. Events are raised using the
    raiseEvent() operation.
    """

    def __init__(self, files):
        """
        Initialize the event manager.
        """
        self.m_handler = {}
        self.m_files = files
        self.m_filewatch = FileWatch(files)

    def configure(self, config):
        """
        Configure the event manager.
        """
        self.m_enabled = config.get('RaiseEvents', 1)
        self.m_filewatch.clear()

    def reload(self):
        """
        Reload the event handler.
        """
        if not self.m_enabled or not self.m_filewatch.modified():
            return
        events = []
        for fi in self.m_files:
            try:
                module = loader.load(fi)
            except:
                logger.error("""** Error loading events **
                        Filename: %s
                        %s
                        """, (fi, getBacktrace()))
                self.m_filewatch.clear(fi)
                continue
            if not module:
                continue
            evlist = []
            for symbol in dir(module):
                attr = getattr(module, symbol)
                if type(attr) == types.ClassType and issubclass(attr, Event) \
                        and attr is not Event:
                    evlist.append(attr)
            if evlist:
                evlist.sort(hiercmp)
                events.append(evlist[0]())
        self.m_handler = {}
        for ev in events:
            self.install(ev)

    def install(self, event):
        """
        Install an event handler. The event argument must be an Event instance.
        """
        if not isinstance(event, Event):
            raise TypeError, 'Expecting Event instance.'
        for name in dir(event):
            method = getattr(event, name)
            if callable(method) and not name.startswith('__'):
                try:
                    self.m_handler[name].append(method)
                except KeyError:
                    self.m_handler[name] = [method]

    def raiseEvent(self, name):
        """
        Raise the event identified by NAME. If a handler is registered for the
        event is is run.
        """
        if not self.m_enabled or not self.m_handler.has_key(name):
            return
        for ha in self.m_handler[name]:
            ha()
