# vi: ts=4 sw=4 et
#
# form.py: form parsing
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: form.py,v 1.14 2003/03/24 11:44:08 geertj Exp $

import draco

from draco.date import Date
from draco.namespace import DictNamespace
from draco.rewrite import Rewriter
from draco.request import FileUpload


class Field:
    """
    Base class of all form fields.
    """

    def __init__(self, name, default=None, absent=None, illegal=None):
        self.name = name
        self.default = default
        self.absent = absent
        self.illegal = illegal

    def parse(self, value):
        return value


class TextField(Field):
    """
    Verify that the field is a string.
    """
    def __init__(self, name, default=None, absent=None, illegal=None,
                    minlen=None, maxlen=None):
        Field.__init__(self, name, default, absent, illegal)
        self.minlen = minlen
        self.maxlen = maxlen

    def parse(self, value):
        if not isinstance(value, str):
            raise ValueError
        if self.minlen is not None and len(value) < self.minlen:
            raise ValueError
        if self.maxlen is not None and len(value) > self.maxlen:
            raise ValueError
        return value


class StringField(TextField):
    """
    Verify that the field is a single line string without control
    characters. The value that is returned is a string with leading and
    trailing whitespace stripped off.
    """

    def parse(self, value):
        if not isinstance(value, str):
            raise ValueError
        value = value.strip()
        if self.minlen is not None and len(value) < self.minlen:
            raise ValueError
        if self.maxlen is not None and len(value) > self.maxlen:
            raise ValueError
        for ch in value:
            if ord(ch) < 0x20:
                raise ValueError
        return value


class AsciiField(TextField):
    """
    Verify that the field is a single line, ascii string without control
    characters. The value that is returned is a string with leading and
    trailing whitespace stripped off.
    """

    def parse(self, value):
        if not isinstance(value, str):
            raise ValueError
        value = value.strip()
        if self.minlen is not None and len(value) < self.minlen:
            raise ValueError
        if self.maxlen is not None and len(value) > self.maxlen:
            raise ValueError
        for ch in value:
            if not 0x20 <= ord(ch) < 0x80:
                raise ValueError
        return value


class IntField(Field):
    """
    Verify that the field contains a valid integer. The returned value
    is a Python integer.
    """

    def __init__(self, name, default=None, absent=None, illegal=None,
                    minval=None, maxval=None):
        Field.__init__(self, name, default, absent, illegal)
        self.minval = minval
        self.maxval = maxval

    def parse(self, value):
        if not isinstance(value, str):
            raise ValueError
        value = int(value)
        if self.minval is not None and value < self.minval:
            raise ValueError
        if self.maxval is not None and value > self.maxval:
            raise ValueError
        return value


class FloatField(Field):
    """
    Verify that the field contains a valid floating point number. The
    returned value is a Python float.
    """

    def __init__(self, name, default=None, absent=None, illegal=None,
                    minval=None, maxval=None):
        Field.__init__(self, name, default, absent, illegal)
        self.minval = minval
        self.maxval = maxval

    def parse(self, value):
        if not isinstance(value, str):
            raise ValueError
        value = float(value)
        if self.minval is not None and value < self.minval:
            raise ValueError
        if self.maxval is not None and value > self.maxval:
            raise ValueError
        return value


class DateField(Field):
    """
    Verify that the field contains a date in a supported format. The
    return value is a Date instance.
    """

    def parse(self, value):
        if not isinstance(value, str):
            raise ValueError
        return Date(value)


class EnumField(Field):
    """
    Verify that the field contains a string value from a predefined set
    of possible string values. The returned value is a string containing
    the selected value.
    """

    def __init__(self, name, values, default=None, absent=None, illegal=None):
        Field.__init__(self, name, default, absent, illegal)
        self.values = values

    def parse(self, value):
        if not isinstance(value, str):
            raise ValueError
        if value not in self.values:
            raise ValueError
        return value


class FileUploadField(Field):
    """
    Verify that the field is a file upload field. The returned value is
    a FileUpload instance.
    """

    def parse(self, value):
        if not isinstance(value, FileUpload):
            raise ValueError
        return value


class FormError(draco.HttpInternalServerError):
    """
    A form was entered incompletely or incorrectly.
    """

    def __init__(self, status='', err_fields=[]):
        self.status = status
        self.err_fields = err_fields

    def __str__(self):
        return self.status


class Form(DictNamespace):
    """
    A form consists of a number of input fields and output variables. The
    input fields are converted to the output variables using the parse()
    function. Form implements a namespace to access the output variables.

    When a field is added using addField(), it becomes an input field and
    an output variable by default. The Form.parse() method uses
    Field.parse() function to convert from input (mod_python types) to
    the output format (Python types).

    By overriding the parse() method, an arbitrary (input field -> output
    variable) mapping can be constructed.
    """

    def __init__(self):
        DictNamespace.__init__(self)
        self.m_fields = []

    def fields(self):
        """
        Return a list containing the form fields.
        """
        return self.m_fields

    def addField(self, field):
        """
        Add a field to the form. The field argument must be a Field instance.
        """
        if not isinstance(field, Field):
            raise TypeError, 'Expecting a Field instance.'
        self.m_fields.append(field)

    def parse(self, args):
        """
        Parse the form, raising a FormError on error.
        """
        for fi in self.m_fields:
            if args.get(fi.name):
                try:
                    value = fi.parse(args[fi.name])
                    self[fi.name] = value
                except ValueError, err:
                    message = fi.illegal or 'U heeft een ongeldige waarde ingevuld.'
                    raise FormError(message, [fi.name])
            elif fi.default is not None:
                self[fi.name] = fi.default
            else:
                message = fi.absent or 'U heeft niet alle verplichte velden ingevuld.'
                raise FormError(message, [fi.name])


def ischecked(value, ref=None):
    """
    Return the string 'checked' if VALUE is true.
    """
    if ref is None:
        if value:
            return 'checked'
    elif isinstance(ref, str):
        if str(value) == str(ref):
            return 'checked'
    elif isinstance(ref, list):
        for re in ref:
            if str(value) == str(re):
                return 'checked'
    return ''


def isselected(value, ref):
    """
    Return the string 'selected' if VAR equals VALUE.
    """
    if isinstance(ref, str):
        if str(value) == str(ref):
            return 'selected'
    elif isinstance(ref, list):
        for re in ref:
            if str(value) == str(re):
                return 'selected'
    return ''


class FormRewriter(Rewriter):
    """
    A Rewriter that can put back form data into html.
    """

    def __init__(self, values, formid=None):
        """
        Create a new form rewriter. It will plug back values from the
        namespace `values' in the form with id `formid'. If no `formid'
        parameter is given, all forms are rewritten.
        """
        self.m_values = values
        self.m_formid = formid
        self.m_form = None
        self.m_select = None

    def form(self, tag):
        """
        Remember the "id" attribute.
        """
        try:
            self.m_form = tag.attrs['id']
        except KeyError:
            self.m_form = None

    def input(self, tag):
        """
        Set the "value" attribute of <input> tags.
        """
        if self.m_formid and self.m_form != self.m_formid:
            return
        try:
            type = tag.attrs['type']
            if type == 'checkbox':
                if ischecked(tag.attrs['value'], self.m_values[tag.attrs['name']]):
                    tag.attrs['checked'] = None
                else:
                    del tag.attrs['checked']
            elif type == 'radio':
                if ischecked(tag.attrs['value'], self.m_values[tag.attrs['name']]):
                    tag.attrs['checked'] = None
                else:
                    del tag.attrs['checked']
            elif type in ('text', 'hidden', 'password'):
                value = str(self.m_values[tag.attrs['name']])
                tag.attrs['value'] = value.encode('attr')
        except KeyError:
            pass

    def select(self, tag):
        """
        Remember the "name" attribute.
        """
        if self.m_formid and self.m_form != self.m_formid:
            return
        self.m_select = tag.attrs.get('name')

    def option(self, tag):
        """
        Make an option "selected" if its matches the value of its
        containing <select>.
        """
        if self.m_formid and self.m_form != self.m_formid:
            return
        try:
            if isselected(tag.attrs['value'], self.m_values[self.m_select]):
                tag.attrs['selected'] = None
            else:
                del tag.attrs['selected']
        except KeyError:
            pass

    def textarea(self, tag):
        """
        Set the text directly after a <textarea>.
        """
        if self.m_formid and self.m_form != self.m_formid:
            return
        try:
            value = str(self.m_values[tag.attrs['name'] ])
            tag.appendText(value.encode('html'))
        except KeyError:
            pass
