# vi: ts=4 sw=4 et
#
# handler.py: the Handler and HandlerManager.
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: handler.py,v 1.33 2003/03/10 16:38:37 geertj Exp $

import types
import draco

from draco.loader import loader
from draco.namespace import DictNamespace
from draco.filewatch import FileWatch
from draco.util import hiercmp, getBacktrace


class Handler(DictNamespace):
    """
    Base class for all document handlers.

    To define a new handler, derive from this class and add functions
    with names of the pages to serve.
    """

    AUTH_OK = 0
    AUTH_DENY = 1

    def __init__(self):
        if self.__class__ == Handler:
            raise NotImplementedError, 'Handler should be subclassed.'
        DictNamespace.__init__(self)


class HandlerManager:
    """
    The handler manager loads and caches handler classes on demand. 

    Depends: loader
    """

    def __init__(self):
        """
        Initialize the handler manager.
        """
        self.m_filewatch = FileWatch()
        self.m_cache = {}

    def configure(self, config):
        """
        Configure the handler manager.
        """
        self.m_filewatch.clear()

    def reload(self):
        pass

    def loadHandler(self, directory):
        """
        Load and return a handler class for the directory `directory'.
        """
        mfile = directory + '/__handler__.py'
        if not self.m_filewatch.modified(mfile) and self.m_cache.has_key(mfile):
            return self.m_cache[mfile]
        try:
            module = loader.load(mfile)
        except:
            self.m_filewatch.clear(mfile)
            error = draco.HttpInternalServerError('Error loading handler')
            error.filename = mfile
            error.traceback = getBacktrace()
            raise error
        if not module:
            return
        hlist = []
        for symbol in dir(module):
            attr = getattr(module, symbol)
            if type(attr) == types.ClassType and issubclass(attr, Handler) \
                    and attr is not Handler:
                hlist.append(attr)
        if not hlist:
            return
        hlist.sort(hiercmp)
        handler = hlist[0]
        self.m_cache[mfile] = handler
        return handler
