# vi: ts=4 sw=4 et
#
# image.py: image size lookup and cache
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: image.py,v 1.14 2003/03/10 16:38:37 geertj Exp $

import os
import struct
import stat

# A short technical note about JPEG headers and how to extract the
# image dimensions:
#
# A JFIF header consists of blocks. A block begins with one or more
# '0xff' bytes, followed by a non 0xff byte indicating the type of the
# block. This is then followed by a two byte length for the block.
# We are interested in block of types in the range 0xc0..0xcf, the
# so called SOFn (Start Of Frame n) blocks. From these we can extract
# the image dimensions.


class ImageManager:
    """
    ImageManager looks up and caches the size of image files. This class is
    used by the default <img> rewriter to add "width" and "height" attributes.

    Caching is done purely in memory. This is allright because looking up
    an image size is very cheap and caching it takes only a tiny bit of memory.

    Expiration is done with a simple 2-threshold LRU algorithm.
    """

    def __init__(self, root):
        """
        Create a new ImageManager object.
        """
        self.m_cache = {}
        self.m_serial = 0
        self.m_docroot = root

    def configure(self, config):
        """
        Configure the image manager.
        """
        minsize = config.get('ImageSizeCache', 500)
        self.m_min = minsize
        self.m_max = min(1.2 * minsize, minsize + 100)

    def reload(self):
        pass

    def _lookup(self, fname):
        """
        Determine the image type of FNAME and return its size.
        """
        try:
            file = open(fname)
        except IOError:
            return
        head = file.read(24)
        if len(head) != 24:
            return
        if head[:4] == '\x89PNG':
            # PNG
            check = struct.unpack('>i', head[4:8])[0]
            if check != 0x0d0a1a0a:
                return
            w, h = struct.unpack('>ii', head[16:24])
        elif head[:6] in ('GIF87a', 'GIF89a'):
            # GIF
            w, h = struct.unpack('<HH', head[6:10])
        elif head[:4] == '\xff\xd8\xff\xe0' and head[6:10] == 'JFIF':
            # JPEG
            try:
                file.seek(0)  # Read 0xff next
                size = 2
                type = 0
                while not 0xc0 <= type <= 0xcf:
                    file.seek(size, 1)
                    b = file.read(1)
                    while ord(b) == 0xff: 
                        b = file.read(1)
                    type = ord(b)
                    size = struct.unpack('>H', file.read(2))[0] - 2
                # We are at a SOFn block
                file.seek(1, 1)  # Skip `precision' byte.
                h, w = struct.unpack('>HH', file.read(4))
            except:
                return
        else:
            return
        return w, h

    def lookup(self, image):
        """
        Lookup and cache the size of the image file IMAGE. IMAGE is relative
        to the document root. If successfull, the tuple (width, height) is
        returned, otherwise None.
        """
        fname = self.m_docroot + '/' + image
        try:
            st = os.stat(fname)
        except OSError:
            return
        if not stat.S_ISREG(st.st_mode):
            return
        mtime = st.st_mtime
        if self.m_cache.has_key(image) and mtime == self.m_cache[image][2]:
            self.m_cache[image][3] = self.m_serial
            self.m_serial += 1
            return tuple(self.m_cache[image][:2])
        size = self._lookup(fname)
        if not size:
            return
        self.m_cache[image] = [size[0], size[1], mtime, self.m_serial]
        self.m_serial += 1
        self.expire()
        return size

    def expire(self):
        """
        Expire the internal cache.
        """
        if len(self.m_cache) < self.m_max:
            return
        expire = []
        for i in self.m_cache.keys():
            expire.append((self.m_cache[i][3], i))
        explist.sort()
        nexpire = self.m_max - self.m_min
        for age, img in expire[:nexpire]:
            del self.m_cache[img]
