# vi: ts=4 sw=4 et
#
# namespace.py: draco namespaces
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: namespace.py,v 1.6 2003/03/17 11:27:19 geertj Exp $


class Namespace:
    """
    Abstract base class for all namespaces in Draco.

    Namespaces are (key, value) mappings identical to Python's dictionaries,
    with the restriction that keys must be strings.

    To implement a Namespace, a derived class has to define at least the
    functions __len__, __getitem__, __setitem__, __delitem__ and items(). All
    other functins have default implementations in terms of these primitives.
    More methods can be defined as an optimization.
    """

    def __init__(self):
        pass

    def __len__(self):
        raise NotImplementedError
        
    def __getitem__(self, name):
        raise NotImplementedError

    def __setitem__(self, name, value):
        raise NotImplementedError

    def __delitem__(self, name):
        raise NotImplementedError

    def items(self):
        raise NotImplementedError

    iteritems = items

    def keys(self):
        return [ i[0] for i in self.items() ]

    iterkeys = __iter__ = keys

    def values(self):
        return [ i[1] for i in self.items() ]

    iteritems = values

    def has_key(self, name):
        try:
            self[name]
            return 1
        except KeyError:
            return 0

    __contains__ = has_key

    def update(self, ns):
        for key,value in ns.items():
            self[key] = value

    def clear(self):
        for key in self.keys():
            del self[key]

    def get(self, key, value=None):
        try:
            return self[key]
        except KeyError:
            return value

    def setdefault(self, key, value=None):
        try:
            return self[key]
        except KeyError:
            self[key] = value
            return value

    def popitem(self, key):
        key, value = self.items()[0]
        del self[key]
        return value

    def copy(self):
        dict = {}
        for key,value in self.items():
            dict[key] = value
        return dict


class DictNamespace(Namespace):
    """
    Concrete namespace using a dictionary as backing store.
    """

    def __init__(self):
        Namespace.__init__(self)
        self.m_dict = {}

    def __len__(self):
        return len(self.m_dict)

    def __getitem__(self, name):
        return self.m_dict[name]

    def __setitem__(self, name, value):
        if type(name) != str:
            raise TypeError, 'Key must be a string.'
        self.m_dict[name] = value

    def __delitem__(self, name):
        del self.m_dict[name]

    def items(self):
        return self.m_dict.items()


class NamespaceBridge:
    """
    A bridge between a Namespace and a concrete realization of it, decoupling
    their types. This is used for example for the database namespaces because
    their type depends on the runtime configuration.
    """

    def __init__(self):
        self.m_namespace = None

    def setNamespace(self, namespace):
        self.m_namespace = namespace

    def __getattr__(self, name):
        return getattr(self.m_namespace, name)
