# vi: ts=4 sw=4 et
#
# request.py: the Request object
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: request.py,v 1.36 2003/03/10 16:38:37 geertj Exp $

import re
import draco
import os.path

from draco.namespace import DictNamespace
from draco.filewatch import FileWatch
from draco.cookie import cookies

from mod_python.util import FieldStorage, Field
from bisect import bisect_right


class RobotSignatures:
    """
    A class that loads and keeps up to date the robot signatures.
    """

    def __init__(self, files):
        """
        Initialize the robot signatures class. FILES must be a sequence
        of signature files to parse.
        """
        self.m_files = files
        self.m_filewatch = FileWatch(files)
        self.m_robots = []

    def reload(self):
        """
        Reload the robot signatures if they have been changed.
        """
        if not self.m_filewatch.modified():
            return
        self.m_robots = []
        for i in self.m_files:
            self.parse(i)
        self.m_robots.sort()

    def parse(self, fname):
        """
        Read robots signatures from file FNAME.
        """
        try:
            fsigs = open(fname)
        except IOError:
            return
        for line in fsigs:
            line = line.strip()
            if not line or line[0] == '#':
                continue
            self.m_robots.append(line.lower())
        fsigs.close()

    def signatures(self):
        """
        Return a sorted list containing the robot signatures.
        """
        return self.m_robots


class FileUpload:
    """
    Adapter class that slightly changes the semantics of mod_python's Field
    class. It imports the file operations from the temporary file to the
    current class so it can be used as a file
    """

    def __init__(self, field):
        self.field = field

    def __getattr__(self, name):
        """
        Aggregate file methods into the current object.
        """
        if hasattr(self.field.file, name):
            attr = getattr(self.field.file, name)
            if callable(attr):
                return attr
        return getattr(self.field, name)


class Request:
    """
    Request provides access to the client request. In addition, is does:

    * User-Agent parsing
    * URI parsing
    * Cookie handling
    """

    def __init__(self):
        """
        Initialize the request object.
        """
        self.m_robots = None

    def configure(self, config):
        """
        Configure the request.
        """
        self.m_docroot = config['DocumentRoot']

    def reload(self, mpreq):
        """
        Reload this request object.
        """
        self.m_mpreq = mpreq
        if self.m_robots is None:
            files = os.path.normpath(draco.cfg_datadir + '/robots.ini'), \
                    os.path.normpath(self.m_docroot + '/robots.ini')
            self.m_robots = RobotSignatures(files)
        self.m_robots.reload()
        self.parseAgent()
        self.parseUri()
        cookies.parseCookies(self)

    re_agent1 = re.compile("([^/(]+)[-/ ]([0-9]+(?:\.[0-9]+)?)")
    re_agent2 = re.compile("([^/(]+)[-/ ]([0-9]+(?:\.[0-9]+)?).*\((.*)\)")

    def parseAgent(self):
        """
        Parse the `User-agent' HTTP header and the decoded result in m_agent. 
        This is mainly used to recognise web spiders, but it could be used to 
        offer browser specific pages.
        """
        self.m_agent = ''
        self.m_agentver = 0.0
        self.m_agentinfo = []
        self.m_isrobot = 0
        mpreq = self.m_mpreq
        if not mpreq.headers_in.has_key('User-Agent'):
            return  
        agent = mpreq.headers_in['User-Agent']
        match = self.re_agent2.match(agent)
        if match:
            name = match.group(1)
            version = match.group(2)
            info = [ i.strip() for i in match.group(3).split(';') ]
            # So called compatible browser? (MSIE)
            if info[0] == 'compatible' and len(info) >= 2:
                match = self.re_agent1.match(info[1])
                if match:
                    name = match.group(1)
                    version = match.group(2)
                    info = info[2:]
            self.m_agent = name
            self.m_agentver = float(version)
            self.m_agentinfo = info
        else:
            match = self.re_agent1.match(agent)
            if match:
                self.m_agent = match.group(1)
                self.m_agentver = float(match.group(2))
        if not self.m_agent:
            return
        # Match the user agent with the robot signatures
        sigs = self.m_robots.signatures()
        agents = [ self.m_agent ] + self.m_agentinfo
        agents = [ i.lower() for i in agents if i ]
        for ag in agents:
            i = bisect_right(sigs, ag)
            if i > 0 and ag.startswith(sigs[i-1]):
                self.m_isrobot = 1
                break

    def parseUri(self):
        """
        The URL is already broken up in filename and pathinfo by apache.
        Extract dir, file, sessionid and convert path_info to a list.
        """
        mpreq = self.m_mpreq
        self.m_path = mpreq.filename
        if self.m_path.startswith(self.m_docroot):
            self.m_relpath = self.m_path[len(self.m_docroot):]
        else:
            self.m_relpath = self.m_path
        self.m_dir = os.path.dirname(self.m_path)
        self.m_reldir = os.path.dirname(self.m_relpath)
        self.m_filename = os.path.basename(self.m_path)
        self.m_pathinfo = []
        if mpreq.path_info:
            path = filter(None, mpreq.path_info.split('/'))
        else:
            path = []
        from session import validSessionId
        if len(path) and validSessionId(path[-1]):
            self.m_sid = path[-1]
            self.m_pathinfo = path[:-1]
        else:
            self.m_sid = None
            self.m_pathinfo = path
        args = DictNamespace()
        fs = FieldStorage(mpreq)
        for name in fs.keys():
            value = fs[name]
            if isinstance(value, str):
                value = str(value)
            elif isinstance(value, Field):
                value = FileUpload(value)
            elif isinstance(value, list):
                for i in range(len(value)):
                    if isinstance(value[i], str):
                        value[i] = str(value[i])
                    elif isinstance(value[i], Field):
                        value[i] = FileUpload(value[i])
            if name:  # mod_python sometimes gives a None key
                args[name] = value
        self.m_args = args

    def method(self):
        """
        Returns HTTP request method (i.e. 'GET', 'HEAD' etc).
        """
        mpreq = self.m_mpreq
        return mpreq.method

    def header(self, name):
        """
        Returns the value of a HTTP header of the request.
        """
        mpreq = self.m_mpreq
        try:
            return mpreq.headers_in[name]
        except KeyError:
            pass

    def agent(self):
        """
        Returns an abridged version of the User-Agent header.If no User-Agent
        header was given in the request, an empty string is returned.
        """
        return self.m_agent

    def agentVersion(self):
        """
        Returns the agent version as parsed from the User-Agent header as a
        floating point number. If no User-Agent header was provided, or one
        was provided in an unknown format, 0.0 is returned.
        """
        return self.m_agentver

    def agentInfo(self):
        """
        Returns extra information as parsed from the User-Agent header as a
        list of strings. If no User-Agent header was provided, or one was
        provided in an unknown format, an empty list is returned.
        """
        return self.m_agentinfo

    def isrobot(self):
        """
        Return nonzero if the User-Agent header matches a robot signature.
        """
        return self.m_isrobot

    def path(self):
        """
        Returns the full pathname of the requested template.
        """
        return self.m_path

    def relpath(self):
        """
        Same as path(), except relative to the document root.
        """
        return self.m_relpath

    def dir(self):
        """
        Returns the directory of the requested template.
        """
        return self.m_dir

    def reldir(self):
        """
        Same as dir(), except relative to the document root.
        """
        return self.m_reldir

    def filename(self):
        """
        Returns the name or the requested template.
        """
        return self.m_filename

    def pathinfo(self):
        """
        Returns that path after the page (CGI's PATH_INFO). The result is returned
        as a list consisting of path entries. The list has no empty items.
        """
        return self.m_pathinfo

    def sessionid(self):
        """
        Returns the session if from the URI, if one was passed.
        """
        return self.m_sid

    def args(self):
        """
        Returns the form arguments as a mod_python FieldStorage class 
        (extracted from the request URI).
        """
        return self.m_args

    def mprequest(self):
        """
        Returns mod_python's request object. This function should only be
        used if there is a function in the mod_python API that is not
        available in the Draco API.
        """
        return self.m_mpreq
