# vi: ts=4 sw=4 et
#
# response.py: the Reponse object.
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: response.py,v 1.39 2003/03/11 09:17:42 geertj Exp $

import urlparse
from cStringIO import StringIO

import draco
from draco.config import config
from draco.cookie import cookies


class Response:
    """
    This object contains all functionality related to the HTTP response.
    """

    def __init__(self):
        """
        Initialize this Response object. By default, the Content-Type
        HTTP header is set to 'text/html'.
        """

    def configure(self, config):
        """
        Configure the response object.
        """
        self.m_compress = config.get('Compression', 1)
        self.m_encoding = config.get('Encoding')
        self.m_appendsid = config.get('SessionAppendUrl', 1)

    def reload(self, mpreq):
        """
        Reload the response object.
        """
        self.m_mpreq = mpreq
        self.m_ctype = 'text/html'
        self.m_modified = None
        self.m_headers = {}
        self.m_headersent = 0
        self.m_buffer = StringIO()
        self.m_buffered = 0

    def setBuffering(self, buffer):
        """
        Buffer the response if `buffer' is nonzero.
        """
        if self.m_headersent or self.m_buffer.tell():
            return
        self.m_buffered = buffer

    def setEncoding(self, encoding):
        """
        Set the response encoding to `encoding'. If the response is a regular
        string, the response is assumed to be already in this encoding and no
        encoding is performed. If the reponse is a unicode string, it is
        encoded to a byte stream using this encoding.
        """
        self.m_encoding = encoding

    def setCompression(self, compress):
        """
        Enable transparent compression if `compress' is nonzero, disable it
        otherwise.
        """
        self.m_compress = compress

    def setContentType(self, ctype):
        """
        Set the Content-Type for the response.
        """
        self.m_ctype = ctype

    def setHeader(self, name, value):
        """
        Set the http header `name' to `value'. If a http header with the
        same name is already set, it is cleared.
        """
        self.m_headers[name] = [value]

    def addHeader(self, name, value):
        """
        Add an arbitrary http header `name' with value `value'.
        """
        try:
            self.m_headers[name].append(value)
        except KeyError:
            self.m_headers[name] = [value]

    def setModified(self, modified):
        """
        This resource was last modified on MODIFIED, which must be a
        DateTime instance.
        """
        if self.m_modified:
            self.m_modified = max(self.m_modified, modified)
        else:
            self.m_modified = modified

    def contentEncodingSupported(self, encoding):
        """
        Return nonzero if the Content-Encoding `encoding' is supported by
        the remote client. Note that this is the content encoding and has
        nothing to do with the character set encoding selected by
        setEncoding() above.
        """
        mpreq = self.m_mpreq
        try:
            accept = mpreq.headers_in['Accept-Encoding']
        except KeyError:
            return 0
        encodings = [ i.strip() for i in accept.split(',') ]
        if accept == '*' or encoding in encodings:
            return 1
        return 0

    def sendHeader(self):
        """
        Send the http header to the client. This is called automatically from
        write() or flush(), depeding on whether buffering is enabled or not.
        """
        cookies.setCookies(self)
        if self.m_modified:
            self.setHeader('Last-Modified', self.m_modified.format('rfc1123'))
        if self.m_encoding and self.m_ctype.startswith('text/'):
            ctype = '%s; charset=%s' % (self.m_ctype, self.m_encoding)
        else:
            ctype = self.m_ctype
        self.m_mpreq.content_type = ctype
        for name in self.m_headers:
            for value in self.m_headers[name]:
                self.m_mpreq.headers_out.add(name, value)
        self.m_mpreq.send_http_header()
        self.m_headersent = 1

    def write(self, buf):
        """
        Add the string `buf' to the response. If the response is in
        buffered mode, `buf' appended to the buffer and this call returns.
        If not, the string is written directly to the client.
        """
        if type(buf) == unicode:
            if not self.m_encoding:
                self.m_encoding = 'utf-8'
            buf = buf.encode(self.m_encoding)
        if self.m_buffered:
            self.m_buffer.write(buf)
            return
        if not self.m_headersent:
            self.sendHeader()
        try:
            self.m_mpreq.write(buf)
        except IOError:
            # Client closed connection. This happens all the time.
            pass

    def flush(self):
        """
        Flush the buffered response.
        """
        if not self.m_buffered:
            return
        if not self.m_encoding:
            self.m_encoding = 'iso-8859-1'
        buffer = self.m_buffer.getvalue()
        if self.m_compress and self.m_ctype.startswith('text/') and \
                self.contentEncodingSupported('deflate'):
            buffer = buffer.encode('zlib')[2:-4]  # Skip zlib head/tail
            self.setHeader('Content-Encoding', 'deflate')
        self.setHeader('Content-Length', str(len(buffer)))
        self.sendHeader()
        try:
            self.m_mpreq.write(buffer)
        except IOError:
            pass

    def redirect(self, path, protocol=None, host=None):
        """
        Redirect the client to URL. If URL is an absolute URL, no extra
        processing is done and the client is redirected directly.
        If URL is a relative URL, the session id is appended to the URL.
        PROTOCOL specifies the protocol to use. If it is not given, the
        the current protocol is used.
        """
        # Use err_headers_out because with a 302 apache uses those.
        cookies.setCookies(self)
        for name in self.m_headers:
            for value in self.m_headers[name]:
                self.m_mpreq.err_headers_out.add(name, value)
        from draco.server import server
        if protocol is None:
            httpsport = config.get('HttpsPort', 443)
            if server.port() == httpsport:
                protocol = 'https'
            else:
                protocol = 'http'
        if host:
            absurl = protocol + '://' + host + path
        else:
            if not path.startswith('/'):
                from draco.request import request
                path = urlparse.urljoin(request.relpath(), path)
            if self.m_appendsid:
                from draco.session import addSessionId
                path = addSessionId(path)
            absurl = protocol + '://' + server.hostname() + path
        self.m_mpreq.err_headers_out.add('Location', absurl)
        raise draco.HttpFound

    def exit(self, code=200):
        """
        Return to the user with HTTP exit code CODE.
        """
        cookies.setCookies(self)
        req = self.m_mpreq
        for name in self.m_headers:
            for value in self.m_headers[name]:
                if code < 300:
                    req.headers_out.add(name, value)
                else:
                    req.err_headers_out.add(name, value)
        exc = draco.HttpResponse()
        exc.code = code
        raise exc
