# vi: ts=4 sw=4 et
#
# rewrite.py: the tag rewrite subsystem
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: rewrite.py,v 1.32 2003/03/24 09:31:29 geertj Exp $

import types

from draco.logger import logger
from draco.loader import loader
from draco.util import hiercmp, getBacktrace
from draco.filewatch import FileWatch


class Rewriter:
    """
    Base class for all tag rewriters.

    To implement a tag rewriter, derive from this class and define
    methods with names of the tags to rewrite.

    A rewrite methods can change the attributes of an html tag. It gets
    passed a dictionary containing the current attributes of the tag.
    The attributes are updated by updating the dictionary.
    """

    def install(self):
        from draco.rewrite import rewritemanager
        rewritemanager.install(self)


class RewriteManager:
    """
    The RewriteManager loads ag rewriter from the files passed into the
    constructor and keeps them up to date. New rewriters can be installed
    with install(). The list of rewriters can be obtained using rewriter().
    """

    def __init__(self, files):
        self.m_files = files
        self.m_filewatch = FileWatch(files)

    def configure(self, config):
        self.m_rwbase = self.m_rewriters = []
        self.m_enabled = config.get('TagRewriting', 1)
        self.m_filewatch.clear()

    def reload(self):
        """
        Reload the tag rewriter. This loads the system rewriter and possibly
        a rewriter defined in __rewrite__.py.
        """
        if not self.m_enabled:
            return
        if not self.m_filewatch.modified():
            self.m_rewriters = self.m_rwbase
            return
        rewriters = []
        for fi in self.m_files:
            try:
                module = loader.load(fi)
            except:
                logger.error("""** Error loading a rewriter **
                        Filename: %s
                        %s
                        """, (fi, getBacktrace()))
                self.m_filewatch.clear(fi)
                continue
            if not module:
                continue
            rwlist = []
            for name in dir(module):
                attr = getattr(module, name)
                if type(attr) == types.ClassType and issubclass(attr, Rewriter) \
                        and attr is not Rewriter:
                    rwlist.append(attr)
            if rwlist:
                rwlist.sort(hiercmp)
                rewriters.append(rwlist[0]())
        self.m_rwbase = self.m_rewriters = rewriters

    def install(self, rewriter):
        """
        Install a rewriter. The rewriter argument must be a Rewriter instance.
        """
        if not isinstance(rewriter, Rewriter):
            raise TypeError, 'Expecting Rewriter instance.'
        if self.m_rewriters is self.m_rwbase:
            self.m_rewriters = self.m_rwbase[:]
        self.m_rewriters.append(rewriter)

    def rewriters(self):
        """
        Returns the current tag rewriters.
        """
        return self.m_rewriters
