# vi: ts=4 sw=4 et
#
# server.py: the Server object
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: server.py,v 1.21 2003/03/17 14:35:57 geertj Exp $

from draco.namespace import NamespaceBridge
from draco.database import database


class ServerNamespace(NamespaceBridge):
    """
    The "server" namespace. This is a namespace that persists forever
    on the server.
    """

    def setScope(self, scope):
        namespace = database.createNamespace('draco_server_data', \
                        'serverid', 1, scope)
        self.setNamespace(namespace)


class Server(ServerNamespace):
    """
    This class can be used to get some information about the webserver
    serving the current request.
    """

    def configure(self, config):
        pass

    def reload(self, mpreq):
        """
        Reloads the server object.
        """
        self.m_mpreq = mpreq
        self.setScope('__system__')

    def __nonzero__(self):
        """
        Check if the server namespaces are available.
        """
        return self.available()

    def namespace(self, scope):
        """
        Return a server namespace with scope SCOPE.
        """
        namespace = ServerNamespace()
        namespace.setScope(scope)
        return namespace

    def generation(self):
        """
        Returns the generation for the current child. The generation starts
        at one after a server reset, and is incremented at every reload.
        """
        # Requires my patch for mod_python 2.7.x
        try:
            return self.m_mpreq.server.my_generation
        except AttributeError:
            return None

    def restart(self):
        """
        Returns that last restart time for this server.
        """
        # Requires my patch for mod_python 2.7.x
        try:
            return self.m_mpreq.server.restart_time
        except AttributeError:
            return None

    def childnum(self):
        """
        Returns the child number of the current child.
        """
        return self.m_mpreq.connection.child_num

    def address(self):
        """
        Return the address the web server is listening on.
        """
        return self.m_mpreq.connection.local_addr[0]
        
    def port(self):
        """
        Return the port the web server is listening on.
        """
        return self.m_mpreq.connection.local_addr[1]

    def hostname(self):
        """
        Returns the hostname of the web server.
        """
        return self.m_mpreq.server.server_hostname

    def admin(self):
        """
        Returns the E-mail address of the server administrator.
        """
        return self.m_mpreq.server.server_admin
