# vi: ts=4 sw=4 et
#
# session.py: session management
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: session.py,v 1.49 2003/03/24 08:03:38 geertj Exp $

import md5
import draco

from draco import util
from draco.logger import logger
from draco.config import config
from draco.namespace import NamespaceBridge
from draco.database import database
from draco.request import request
from draco.cookie import cookies
from draco.server import server
from draco.date import DateTime


def validSessionId(sid):
    """
    Return nonzero if SID is a valid session id.
    """
    if len(sid) != 21 or sid[3] != '-' or sid[12] != '-':
        return 0
    try:
        n = [ int(i) for i in (sid[0:3], sid[4:12], sid[13:21]) ]
    except ValueError:
        return 0
    if (n[1] + n[2]) % 999 != n[0]:
        return 0
    return 1


def addSessionId(url):
    """
    Add the session id to URL, but only if it is a relatively
    linked template.
    """
    sid = session.sessionid()
    if not sid:
        return url
    pos = url.rfind('?')
    if pos >= 0:
        new = url[:pos]
    else:
        new = url
    ext = '.' + config.get('Extension', 'dsp')
    istmpl = new.find(ext + '/') >= 0 or new.endswith(ext)
    if new.startswith('http://') or not istmpl:
        return url
    if new[-1] != '/':
        new += '/'
    new += sid
    if pos >= 0:
        new += url[pos:]
    return new


class SessionNamespace(NamespaceBridge):
    """
    The "session" namespace: a namespace is bound to the current session.
    """

    def setScope(self, sessionid, scope):
        namespace = database.createNamespace('draco_session_data',
                        'sessionid', sessionid, scope)
        self.setNamespace(namespace)


class Session(SessionNamespace):
    """
    This class performs session management. Its main functionality is:

    * The default session namespace (scope = '__system__')
    * Load, create and update sessions.
    * Perform session login, logout or set persistency.

    Depends: database, request, cookie
    """

    def configure(self, config):
        """
        Configure the session object.
        """
        self.m_enabled = config.get('SessionManagement', 1)
        self.m_timeout = config.get('SessionTimeout', 7200)
        self.m_ptimeout = config.get('SessionPersistentTimeout', 8640000)
        self.m_checkagent = config.get('SessionCheckAgent', 1)
        self.m_secure = config.get('SessionSecure', 0)
        self.m_httpsport = config.get('HttpsPort', 443)

    def reload(self):
        """
        Setup the session. If we have a session id (either from a cookie
        or the URI), check it and load the session data. If there is no
        session id, a new, empty session is created.
        """
        self.m_new = 0
        self.m_session = None
        self.m_userid = 0
        self.m_persistent = 0
        self.m_agent = self.userHash(request.header('User-Agent') or '')
        self.setScope(None, '__system__')
        if not self.m_enabled or not database:
            return
        cursor = database.cursor()
        # Get session id from the URL or from the cookie.
        pathsid = request.sessionid()
        cookiesid = cookies.get('session-id')
        if not cookiesid or not validSessionId(cookiesid):
            cookiesid = None
        if pathsid and cookiesid:
            if pathsid != cookiesid:
                logger.warn('Different path and cookie session ids')
                logger.warn('Taking the path session id')
            self.m_session = pathsid
        elif pathsid:
            self.m_session = pathsid
        elif cookiesid:
            self.m_session = cookiesid
        else:
            self.newSession()
            return
        # Check if the session exists.
        cursor.execute("""
                select userid, auth_token, persistent, expires, user_agent
                from draco_sessions where sessionid = %s
                """, (self.m_session,))
        if cursor.rowcount == 0:
            logger.warn('Unknown but valid session-id %s' % self.m_session)
            self.newSession()
            return
        result = cursor.fetchone()
        self.m_userid = result[0]
        self.m_authtoken = result[1]
        self.m_persistent = result[2]
        self.m_expires = DateTime(result[3])
        # Check if user-agent matches
        if self.m_checkagent and self.m_agent != result[4]:
            logger.warn('Unmatched user agent for session %s' % self.m_session)
            self.newSession()
            return
        if self.m_userid:
            # Check if this a secure session.
            if self.m_secure and server.port() != self.m_httpsport:
                logger.warn('Session %s is not secure' % self.m_session)
                if cookies.get('auth-token') == self.m_authtoken:
                    logger.warn('Unsecure session provided auth token!')
                    self.logout()
                self.newSession()
                return
            # Check if authentication token matches
            if not cookies.get('auth-token'):
                logger.warn('Missing auth token for session %s' % self.m_session)
                self.newSession()
                return
            if cookies['auth-token'] != self.m_authtoken:
                logger.warn('Incorrect auth token for session %s (%s)' % \
                    (self.m_session, cookies['auth-token']))
                self.newSession()
                return
        self.updateSession()

    def generateSessionId(self):
        """
        Generate a new session-id. A simple checksum is used on the sessionid.
        """
        n = (util.randInt() & 0x3ffffff, util.randInt() & 0x03ffffff)
        sid = "%03u-%08u-%08u" % ((n[0] + n[1]) % 999, n[0], n[1])
        return sid

    def generateAuthToken(self):
        """
        Create a new authentication token.
        """
        tok = '%010u-%010u' % (util.randInt(), util.randInt())
        return tok

    def userHash(self, str):
        """
        Return a 32 byte hash of a user-agent string.
        """
        md = md5.new()
        md.update(str)
        return md.hexdigest().upper()

    def newSession(self):
        """
        Creates a new, empty session. For robots (as defined in the config 
        file `robots.conf'), we never create a session!
        """
        if request.isrobot():
            self.m_session = None
            return
        self.m_new = 1
        self.m_userid = None
        self.m_session = self.generateSessionId()
        timeout = DateTime()
        timeout.addTime(0, 0, self.m_timeout)
        cursor = database.cursor()
        cursor.execute("""
                insert into draco_sessions 
                (sessionid, user_agent, create_date, last_used, expires)
                values (%s, %s, NOW(), NOW(), %s)
                """, (self.m_session, self.m_agent, timeout.format('iso8601')))
        database.commit()
        cookies['session-id'] = self.m_session
        cookies.setExpires('session-id', timeout)
        self.setScope(self.m_session, '__system__')

    def updateSession(self):
        """
        Updates the session's expiration date.
        """
        timeout = DateTime()
        if self.persistent():
            timeout.addTime(0, 0, self.m_ptimeout)
        else:
            timeout.addTime(0, 0, self.m_timeout)
        cursor = database.cursor()
        cursor.execute("""
                update draco_sessions set last_used = NOW(), expires = %s
                where sessionid = %s
                """, (timeout.format('iso8601'), self.m_session))
        database.commit()
        cookies['session-id'] = self.m_session
        cookies.setExpires('session-id', timeout)
        if self.m_userid:
            cookies['auth-token'] = self.m_authtoken
            cookies.setExpires('auth-token', timeout)
            cookies.setSecure('auth-token', self.m_secure)
        self.setScope(self.m_session, '__system__')

    def login(self, userid):
        """
        Log in the user with USERID to the current session.
        """
        if not self:
            raise draco.HttpInternalServerError, 'Session not available.'
        self.m_userid = userid
        self.m_authtoken = self.generateAuthToken()
        cursor = database.cursor()
        cursor.execute("""
                update draco_sessions set userid = %s, auth_token = %s
                where sessionid = %s
                """, (userid, self.m_authtoken, self.m_session))
        database.commit()
        timeout = DateTime()
        timeout.addTime(0, 0, self.m_timeout)
        cookies['auth-token'] = self.m_authtoken
        cookies.setExpires('auth-token', timeout)
        cookies.setSecure('auth-token', self.m_secure)

    def logout(self):
        """
        Log out a user. This means we reset the `userid' and `auth_token' field
        for the current session. We also clear the persistent flag and set the
        expiration time to the default for non persistent sessions (sessions that
        are not logged in cannot be persistent).
        """
        if not self:
            raise draco.HttpInternalServerError, 'Session not available.'
        if not self.m_userid:
            return
        self.m_userid = None
        timeout = DateTime()
        timeout.addTime(0, 0, self.m_timeout)
        cursor = database.cursor()
        cursor.execute("""
                update draco_sessions 
                set userid = 0, auth_token = "", persistent = 0, expires = %s
                where sessionid = %s
                """, (timeout.format('iso8601'), self.m_session))
        database.commit()
        self.m_persistent = 0
        cookies.setExpires('session-id', timeout)
        del cookies['auth-token']

    def setPersistent(self, persistent=1):
        """
        Make this is persistent session (kept between browser sessions).
        Only logged in sessions (userid() != None) can be persisent.
        """
        if not self:
            raise draco.HttpInternalServerError, 'Session not available.'
        if not self.m_userid:
            return
        persistent = persistent and 1 or 0
        self.m_persistent = persistent
        timeout = DateTime()
        if persistent:
            timeout.addTime(0, 0, self.m_ptimeout)
        else:
            timeout.addTime(0, 0, self.m_timeout)
        cursor = database.cursor()
        cursor.execute("""
                update draco_sessions set persistent = %s, expires = %s
                where sessionid = %s
                """, (persistent, timeout.format('iso8601'), self.m_session))
        database.commit()
        cookies.setExpires('session-id', timeout)
        cookies.setExpires('auth-token', timeout)
        cookies.setSecure('auth-token', self.m_secure)

    def __nonzero__(self):
        """
        Are the session namespaces available?
        """
        return self.available()

    def namespace(self, scope):
        """
        Return the namespace with scope SCOPE.
        """
        namespace = SessionNamespace()
        namespace.setScope(self.m_session, scope)
        return namespace

    def sessionid(self):
        """
        Returns the session id of the current session or None if there
        is no session.
        """
        return self.m_session

    def enabled(self):
        """
        Return nonzero if session management is enabled.
        """
        return self.m_enabled

    def userid(self):
        """
        Returns the user id of the current session or None if the session
        is not logged in.
        """
        return self.m_userid

    def persistent(self):
        """
        Is this a persistent session?
        """
        return self.m_persistent

    def isnew(self):
        """
        Is this a newly created session?
        """
        return self.m_new
