# vi: ts=4 sw=4 et
#
# task.py: task scheduling
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2003 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: task.py,v 1.22 2003/03/18 13:54:01 geertj Exp $

import os
import sys
import types
from bisect import bisect_left, insort_right

import draco
from draco.logger import logger
from draco.config import config
from draco.loader import loader
from draco.database import database, dbapi
from draco.date import DateTime
from draco.filewatch import FileWatch
from draco.util import getBacktrace


class Task:
    """
    Base class for all scheduled tasks in Draco.

    Tasks are scheduled using their .schedule() function.
    Derived class must implement the run() worker function.
    """

    def __init__(self):
        """
        Construct a new Task object.
        """
        self.m_scheduled = 0

    def schedule(self, minute=None, hour=None, dow=None):
        """
        Schedule this task at HOUR:MINUTE of weekday DOW. Every one of the
        parameters can be either an integer, a tuple of integers or None.
        If an integer is passed, the task is run at the specified time.
        If a tuple of integers is passed, the task is run at every time
        specified in the tuple. If None is passed, this is equivalent to a
        tuple from 0 to the highest value possible for that parameter.
        HOUR, MINUTE and DOW all start from zero. A zero DOW means monday.
        """
        if minute is None:
            minute = range(60)
        elif type(minute) in (int, float, long):
            minute = [ int(minute) ]
        else:
            minute = list(minute)
            minute.sort()
        if minute[0] < 0 or minute[-1] >= 60:
            raise ValueError, 'Illegal minute parameter.'
        self.m_minute = minute
        if hour is None:
            hour = range(24)
        elif type(hour) in (int, float, long):
            hour = [ int(hour) ]
        else:
            hour = list(hour)
            hour.sort()
        if hour[0] < 0 or hour[-1] >= 24:
            raise ValueError, 'Illegal hour parameter.'
        self.m_hour = hour
        if dow is None:
            dow = range(7)
        elif type(dow) in (int, float, long):
            dow = [ int(dow) ]
        else:
            dow = list(dow)
            dow.sort()
        if dow[0] < 0 or dow[-1] >= 7:
            raise ValueError, 'Illegal day of week.'
        self.m_dow = dow
        self.m_scheduled = 1

    def nextrun(self, ref):
        """
        Calculate the time > ref this task is to be run again.
        """
        if not self.m_scheduled:
            return
        dt = ref.copy()
        dt.addTime(0, 1, -dt.second)  # next minute
        minute = dt.minute
        i = bisect_left(self.m_minute, minute)
        try:
            dm = self.m_minute[i] - minute
        except IndexError:
            dm = self.m_minute[0] - minute + 60
        dt.addTime(0, dm, 0)
        hour = dt.hour
        i = bisect_left(self.m_hour, hour)
        try:
            dh = self.m_hour[i] - hour
        except IndexError:
            dh = self.m_hour[0] - hour + 24
        dt.addTime(dh, 0, 0)
        dow = dt.weekday
        i = bisect_left(self.m_dow, dow)
        try:
            dd = self.m_dow[i] - dow
        except IndexError:
            dd = self.m_dow[0] - dow + 7
        dt.addDate(0, 0, dd)
        return dt

    def name(self):
        """
        Return the name of the task.
        """
        return self.__class__.__name__

    def stub(self):
        """
        Run the task in an external environment.
        """
        os.spawnl(os.P_NOWAIT, '%s/bin/python' % sys.exec_prefix, 'python',
                  '%s/dracotask.py' % draco.cfg_bindir, config['DocumentRoot'],
                  self.name())

    def run(self):
        raise NotImplementedError


class TaskQueue:
    """
    A priority queue containing tasks and their next time to run.
    """

    def __init__(self):
        self.m_queue = []

    def __nonzero__(self):
        return len(self.m_queue)

    def add(self, task, ref):
        nextrun = task.nextrun(ref)
        if nextrun is not None:
            insort_right(self.m_queue, (nextrun, task))

    def runnable(self, ref):
        if self.m_queue:
            return self.m_queue[0][0] <= ref
        else:
            return 0

    def pop(self):
        return self.m_queue.pop(0)

    def clear(self):
        self.m_queue = []


class Scheduler:
    """
    The Draco task scheduler. The method runTasks() starts all tasks
    that need running. The tasks are started as an external process in order
    not to hold up the request processing.
    """

    def __init__(self, files):
        self.m_queue = TaskQueue()
        self.m_files = files
        self.m_filewatch = FileWatch(files)

    def configure(self, config):
        self.m_enabled = config.get('TaskScheduling', 1)
        self.m_filewatch.clear()

    def reload(self):
        """
        Reload the task file if is has changed.
        """
        if not self.m_enabled or not self.m_filewatch.modified():
            return
        tasks = {}
        for fi in self.m_files:
            try:
                module = loader.load(fi)
            except:
                logger.error("""** Error loading tasks **
                        Filename: %s
                        %s
                        """, (fi, getBacktrace()))
                self.m_filewatch.clear(fi)
                continue
            if not module:
                continue
            for symbol in dir(module):
                attr = getattr(module, symbol)
                if type(attr) == types.ClassType and issubclass(attr, Task) \
                        and attr is not Task:
                    tasks[symbol] = attr
        now = DateTime()
        self.m_queue.clear()
        for ta in tasks.values():
            self.m_queue.add(ta(), now)

    def runTasks(self):
        """
        Run tasks that need running.
        """
        if not self.m_enabled or not database:
            return
        now = DateTime()
        cursor = database.cursor()
        processed = []
        while self.m_queue.runnable(now):
            nextrun, task = self.m_queue.pop()
            processed.append(task)
            cursor.execute("""
                    select last_run from draco_tasks
                    where name = %s
                    """, (task.name(),))
            if cursor.rowcount:
                lastrun = DateTime(cursor.fetchone()[0])
                if nextrun <= lastrun:
                    continue
                cursor.execute("""
                        update draco_tasks set last_run = %s
                        where name = %s and last_run = %s
                        """, (int(now.time), task.name(), int(lastrun.time)))
                database.commit()
                if not cursor.rowcount:
                    continue
            else:
                try:
                    cursor.execute("""
                            insert into draco_tasks
                            (name, last_run) values (%s, %s)
                            """, (task.name(), int(now.time)))
                    database.commit()
                except dbapi.IntegrityError:
                    database.rollback()
                    continue
            logger.info('Running task %s' % task.name())
            task.stub()
        for task in processed:
            self.m_queue.add(task, now)
