# vi: ts=4 sw=4 et
#
# util.py: various utility functions
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: util.py,v 1.27 2003/03/26 09:35:40 geertj Exp $

import sys
import struct
import traceback
from rexec import RExec


class REvalEnvironment(RExec):
    """
    An environment in which nothing is allowed. 
    It is used to implement reval(), below.
    """

    ok_builtin_modules = ()
    ok_path = ()
    ok_sys_names = ()
    ok_posix_names = ()

    def r_open(self, *args):
        raise IOError, 'Cannot open files.'


def reval(expr):
    """
    Evaluate an expression in a global restricted environment.
    """
    global _reval_environment
    try:
        env = _reval_environment
    except NameError:
        _reval_environment = REvalEnvironment()
        env = _reval_environment
    return env.r_eval(expr)


def hiercmp(c1, c2):
    """
    Compare function that orders the classes c1 and c2 by their heritage.
    """
    if c1 is c2:
        return 0
    elif issubclass(c1, c2):
        return -1
    elif issubclass(c2, c1):
        return +1
    return 0


def attributes(obj):
    """
    Return the non-special attributes of obj.
    """
    attrs = {}
    for at in dir(obj):
        if not at.startswith('__'):
            attrs[at] = getattr(obj, at)
    return attrs


def getBacktrace():
    """
    Return a backtrace from the last exception that occured.
    """
    type, value, trace = sys.exc_info()
    mesg = traceback.format_exception(type, value, trace)
    trace = None  # See python docs
    return ''.join(mesg)


def randInt():
    """
    Returns a 32-bits random integer, extraced from /dev/random.
    """
    f_rnd = open("/dev/urandom", "r")
    buf = f_rnd.read(4)
    i = struct.unpack("i", buf)[0]
    f_rnd.close()
    return i


def reindent(s, indent='', strip=1):
    """
    Re-indent a multiline string. Python's docstring convention is used
    for de-indenting.
    """
    if not s:
        return ''
    lines = s.expandtabs().splitlines()
    level = 0
    for li in lines[1:]:
        if li and not li.isspace():
            while li[level].isspace():
                level += 1
            break
    lines = [ lines[0].strip() ] + \
            [ li[level:].rstrip() for li in lines[1:] ]
    if strip:
        for lo in range(len(lines)):
            if lines[lo] and not lines[lo].isspace():
                break
        for hi in range(len(lines),0,-1):
            if lines[hi-1] and not lines[hi-1].isspace():
                break
    else:
        lo = 0
        hi = len(lines)
    if indent:
        for i in range(max(lo, 1), hi):
            lines[i] = indent + lines[i]
    return '\n'.join(lines[lo:hi])
