# vi: ts=4 sw=4 et
#
# __tasks__.py: system tasks
#
# This file is part of Draco, a dynamic web content system in Python.
# Draco is copyright (C) 1999-2002 Geert Jansen <geert@boskant.nl>.
# You can freely use and distribute this software under the terms of the
# GNU General Public License, version 2. See the file "LICENSE" included
# with this package for the exact licensing terms.
#
# $Id: __tasks__.py,v 1.5 2003/03/19 11:20:30 geertj Exp $

import cPickle

from draco.task import Task
from draco.logger import logger
from draco.database import database, dbapi
from draco.date import DateTime


class ExpireSessions(Task):
    """
    Remove expired sessions and session data from the system.
    """

    def __init__(self):
        """
        This task is scheduled every hour.
        """
        Task.__init__(self)
        self.schedule(minute=5)

    def run(self):
        """
        Expire the sessions.
        """
        logger.info('Expiring sessions')
        now = DateTime()
        cursor = database.cursor()
        cursor.execute("""
                select sessionid from draco_sessions
                where expires < %s
                """, (now,))
        sessions = [ i[0] for i in cursor.fetchall() ]
        for id in sessions:
            cursor.execute("""
                    delete from draco_sessions    
                    where sessionid = %s
                    """, (id,))
            cursor.execute("""
                    delete from draco_session_data
                    where sessionid = %s
                    """, (id,))
            database.commit()
        logger.info('Succesfully expired %d sessions' % len(sessions))


class UpgradeFrom993:
    """
    Update table definitions from Draco 0.99.3.
    """

    tables = (('draco_server_data', 'serverid'),
              ('draco_application_data', 'generation'),
              ('draco_session_data', 'sessionid'),
              ('draco_user_data', 'userid'))

    def run(self):
        """
        Upgrade from 0.99.3.
        """
        logger.info('Upgrading from Draco 0.99.3')
        logger.debug('Updating draco_tasks table')
        cursor = database.cursor()
        cursor.execute("""
                alter table draco_tasks
                add column tmp integer not null
                """)
        cursor.execute("""
                update draco_tasks
                set tmp = UNIX_TIMESTAMP(last_run)
                """)
        cursor.execute("""
                alter table draco_tasks
                drop column last_run
                """)
        cursor.execute("""
                alter table draco_tasks
                change tmp last_run integer not null
                """)
        database.commit()
        logger.debug('Updating draco_server_data table')
        cursor.execute("""
                alter table draco_server_data
                add column serverid integer not null first
                """)
        cursor.execute("""
                update draco_server_data
                set serverid = 1
                """)
        database.commit()
        for table, key in self.tables:
            logger.debug('Converting data for %s' % table)
            cursor.execute("""
                    select %s, scope, name, value
                    from %s
                    """ % (key, table))
            count = 0
            cursor2 = database.connection().cursor()
            while 1:
                row = cursor.fetchone()
                if not row:
                    break
                try:
                    value = eval(row[3])
                    value = cPickle.dumps(value)
                except:
                    logger.error('Could not convert value %s:%s:%s' % tuple(row[:3]))
                    continue
                cursor2.execute("""
                        update %s
                        set value = %%s
                        where %s = %%s and scope = %%s and name = %%s
                        """ % (table, key),
                              (dbapi.Binary(value), row[0], row[1], row[2]))
                count += 1
            logger.debug('Converted %d rows' % count)
            database.commit()
        logger.info('Upgrade done!')
